from collections import (
    defaultdict,
)

from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
)
from django.db.models import (
    ForeignKey,
)

from factory_bo.storages import (
    ExistingFactoryStorage,
)


class Command(BaseCommand):
    """
    Команда для сбора оставшихся сгенерированных фабрик по проектам
    """

    def handle(self, *args, **options):
        existing_factory_storage = ExistingFactoryStorage()

        model_default_factory_map = (
            existing_factory_storage.get_model_default_factory_map()
        )

        factory_classes = defaultdict(set)

        for existing_factory in model_default_factory_map.values():
            factory_classes[existing_factory.module_path].add(existing_factory)

        project = 'web-bb-salary'
        file_path = (
            f'{settings.FACTORY_BO__FIXTURES_DIR_PATH}/'
            f'{project.replace("-", "_")}_wrong_fk_ids.csv'
        )
        with open(file_path, 'a+') as f:
            for module_path, existing_factories in factory_classes.items():
                if project not in module_path:
                    continue

                sorted_existing_factories = (
                    sorted(
                        existing_factories,
                        key=lambda item: item.factory_class_name,
                    )
                )

                for existing_factory in sorted_existing_factories:
                    model = existing_factory.factory_class._meta.model
                    get_or_create_fields = (
                        existing_factory.factory_class.get_django_get_or_create()
                    )

                    model.objects = model._base_manager

                    wrong_fks = set()

                    for field in model._meta.fields:
                        if (
                            isinstance(field, ForeignKey) and
                            field.attname.replace('_id', '') in get_or_create_fields
                        ):
                            wrong_fks.add(field.attname)

                    if wrong_fks:
                        f.write(
                            f'{existing_factory.factory_class_name}, {";".join(wrong_fks)}\n'
                        )




def init(conf):
    # =============================================================================
    # Настройки factory_bo.
    # =============================================================================
    FIXTURES_DIR_PATH = conf.get(
        'factory_bo',
        'FIXTURES_DIR_PATH'
    )

    # Исключаемые из обработки приложения
    EXCLUDED_APPS = conf.get(
        'factory_bo',
        'EXCLUDED_APPS'
    )
    if EXCLUDED_APPS:
        EXCLUDED_APPS = set(
            map(
                lambda x: x.strip(),
                EXCLUDED_APPS.split(',')
            )
        )
    else:
        EXCLUDED_APPS = set()

    # Исключаемые из обработки модели
    EXCLUDED_MODELS = conf.get(
        'factory_bo',
        'EXCLUDED_MODELS'
    )
    if EXCLUDED_MODELS:
        EXCLUDED_MODELS = set(
            map(
                lambda x: x.strip(),
                EXCLUDED_MODELS.split(',')
            )
        )
    else:
        EXCLUDED_MODELS = set()

    # Поля-идентификаторы полей моделей содержащих ссылки на ту же модель
    PSEUDO_SELF_FK_IDS = conf.get(
        'factory_bo',
        'PSEUDO_SELF_FK_IDS'
    )
    if PSEUDO_SELF_FK_IDS:
        PSEUDO_SELF_FK_IDS = list(
            map(
                lambda x: x.strip(),
                PSEUDO_SELF_FK_IDS.split(',')
            )
        )
    else:
        PSEUDO_SELF_FK_IDS = []

    # Поля-идентификаторы, которые являются псевдо внешними ключами на
    # существующие записи таблиц. Значения представляют из себя строки
    # содержащие: имя поля, разделитель в виде трех подчеркиваний, метка
    # модели (Model._meta.label). Например, ent_id___enterprise.Enterprise
    PSEUDO_FK_IDS = conf.get(
        'factory_bo',
        'PSEUDO_FK_IDS'
    )
    if PSEUDO_FK_IDS:
        PSEUDO_FK_IDS = dict(
            map(
                lambda x: x.strip().split('___'),
                PSEUDO_FK_IDS.split(',')
            )
        )
    else:
        PSEUDO_FK_IDS = {}

    return {
        'FIXTURES_DIR_PATH': FIXTURES_DIR_PATH,
        'EXCLUDED_APPS': EXCLUDED_APPS,
        'EXCLUDED_MODELS': EXCLUDED_MODELS,
        'PSEUDO_SELF_FK_IDS': PSEUDO_SELF_FK_IDS,
        'PSEUDO_FK_IDS': PSEUDO_FK_IDS,
    }



import contextlib
import json
from operator import (
    itemgetter,
)
from typing import (
    Callable,
    Tuple,
)

from django.db import (
    connection,
)
from factory import (
    DjangoModelFactory,
    FactoryError,
)
from factory.base import (
    OptionDefault,
)
from factory.django import (
    DjangoOptions,
)

from factory_bo.enums import (
    FactoryUseTypeEnum,
)


class WebBBOptions(DjangoOptions):
    def _build_default_options(self):
        """
        Добавление новых полей, которые валидны для указания в class Meta фабрик
        """
        options = super()._build_default_options()
        options.extend([
            OptionDefault('auto_updated_fields', ()),
            OptionDefault('allow_use_id_field', False),
            OptionDefault('excluded_fk_fields', ()),
        ])

        return options


class WebBBModelFactory(DjangoModelFactory):
    """
    Переопределенный базовый класс фабрики проекта web_bb
    """
    # Указывает вид использования фабрики. У каждой модели должна быть одна
    # соответствующая дефолтная фабрики и неограниченное количество кастомных,
    # которые могут быть использованы по-необходимости в тестах
    _factory_use_type = FactoryUseTypeEnum.DEFAULT

    # Режим проверки наличия записи модели.
    # По-умолчанию инстанцирование класса фабрики выполняет получение записи из БД, и в случае отсутствия записи
    # вызывается исключение. При отключенной проверке в случае отсутствия записи в БД, выполняется
    # стандартная логика создания записи из get_or_create.
    _check_record_existence = True

    _options_class = WebBBOptions

    @classmethod
    def get_django_get_or_create(cls) -> Tuple[str]:
        """
        Возвращает набор полей django_get_or_create фабрики
        """
        return getattr(
            cls._meta,
            'django_get_or_create',
            ()
        )

    @classmethod
    def get_django_get_or_create_getter(cls) -> Callable:
        """
        Возвращает itemgetter набора полей django_get_or_create фабрики
        """
        django_get_or_create = cls.get_django_get_or_create()

        return itemgetter(*django_get_or_create)

    @classmethod
    def get_factory_use_type(cls):
        """
        Возвращает тип использования фабрики
        """
        return cls._factory_use_type

    @classmethod
    def get_model(cls):
        return cls._meta.model

    @classmethod
    def get_model_label(cls):
        """
        Возвращает лейбел модели фабрики
        """
        return cls._meta.model._meta.label

    @classmethod
    def get_auto_updated_fields(cls):
        """
        Возвращает поля с автогенерируемыми значениями для фабрики
        """
        return getattr(cls._meta, 'auto_updated_fields', ())

    @classmethod
    def get_allow_use_id_field(cls):
        """
        Возвращает признак возможности использования поля ID для выборки
        и создания объектов при инстанцировании фабрик
        """
        return getattr(cls._meta, 'allow_use_id_field', False)

    @classmethod
    def _after_postgeneration(cls, instance, create, results=None):
        super()._after_postgeneration(instance, create, results)

        if cls._meta.allow_use_id_field:
            cls.increase_pk_sequence_value(
                instance._meta.db_table,
                instance._meta.pk.column,
            )

    @staticmethod
    def increase_pk_sequence_value(db_table, pk_column):
        """
        Сдвиг последовательности ID таблицы вперёд
        """
        with connection.cursor() as cursor:
            cursor.execute(f"""
                SELECT setval(
                    pg_get_serial_sequence('{db_table}', '{pk_column}'),
                    max({pk_column})
                ) FROM "{db_table}";
            """)

    @classmethod
    def get_excluded_fk_fields(cls):
        """
        Возвращает набор полей внешних ключей для исключения при
        простановке зависимостей
        """
        return getattr(cls._meta, 'excluded_fk_fields', ())

    @classmethod
    def _get_or_create(cls, model_class, *args, **kwargs):
        """Create an instance of the model through objects.get_or_create."""
        cls.check_record_existence(args, kwargs, model_class)
        return super()._get_or_create(model_class, *args, **kwargs)

    @classmethod
    def check_record_existence(cls, args, kwargs, model_class):
        """Проверка наличия записи модели по указанным параметрам."""
        if '_check_exists' in kwargs:
            arg_check_exists = kwargs.pop('_check_exists')
        else:
            arg_check_exists = True

        manager = cls._get_manager(model_class)
        key_fields = {}

        for field in cls._meta.django_get_or_create:
            if field not in kwargs:
                raise FactoryError(
                    f"django_get_or_create - "
                    f"Unable to find initialization value for '{field}' in factory {cls.__name__}"
                )
            key_fields[field] = kwargs[field]

        is_instance_exists = manager.filter(*args, **key_fields).exists()

        if (cls._check_record_existence and arg_check_exists) and not is_instance_exists:
            str_kwargs = '\n\t\t'.join([f'{k}: {v}' for k, v in kwargs.items()])

            # Создание qs'та с фильтрацией по совпадающим полям, сохранение наименований отличающихся.
            not_matching_fields = []
            qs = manager.values(*key_fields)
            for field_with_value in key_fields.items():
                if qs.filter(field_with_value).exists():
                    qs = qs.filter(field_with_value)
                else:
                    not_matching_fields.append(field_with_value[0])

            partially_matching_records = json.dumps(list(qs), default=str, indent=True, ensure_ascii=False)
            raise Exception(
                f'Не найдена запись модели "{model_class.__name__}"!\n'
                f'Параметры фильтрации:\nargs: {args}\nkwargs:\n\t\t{str_kwargs}\n\n'
                f'Записи, отличающиеся значением поля/ей "{", ".join(not_matching_fields)}":\n'
                f'{partially_matching_records}'
            )

    @classmethod
    @contextlib.contextmanager
    def disable_check_existence(cls):
        """Контекстный менеджер отключения проверки наличия записи модели."""
        cls._check_record_existence = False
        yield
        cls._check_record_existence = True


class DefaultManagerFactory(WebBBModelFactory):
    """
    Базовый класс фабрик с переопределенным менеджером
    """

    @classmethod
    def _get_manager(cls, model_class):
        manager = model_class._base_manager

        return manager

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        old_provider_use = getattr(model_class, 'provider_use', None)

        model_class.provider_use = True

        manager = cls._get_manager(model_class)

        if cls._meta.django_get_or_create:
            obj = cls._get_or_create(model_class, *args, **kwargs)
        else:
            obj = manager.create(*args, **kwargs)

        if old_provider_use:
            model_class.provider_use = old_provider_use
        else:
            del model_class.provider_use

        return obj



from collections import (
    defaultdict,
)

from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
)

from factory_bo.consts import (
    GENERATED_ACTUAL_PREFIX,
    GENERATED_PREFIX,
)
from factory_bo.storages import (
    ExistingFactoryStorage,
)


class Command(BaseCommand):
    """
    Команда для сбора оставшихся сгенерированных фабрик по проектам
    """

    def handle(self, *args, **options):
        existing_factory_storage = ExistingFactoryStorage()

        model_default_factory_map = (
            existing_factory_storage.get_model_default_factory_map()
        )

        factory_classes = defaultdict(set)

        for existing_factory in model_default_factory_map.values():
            factory_classes[existing_factory.module_path].add(existing_factory)

        stage = 'web-bb-salary-etalon'
        project = 'web-bb-core'
        file_path = (
            f'{settings.FACTORY_BO__FIXTURES_DIR_PATH}/'
            f'{project.replace("-", "_")}_need_check_{{target}}_factories.csv'
        )
        with open(file_path.format(target='etalon'), 'a+') as etalon_file:
            etalon_file.write(f'{stage},\n')
            with open(file_path.format(target='generated_actual'), 'a+') as generated_actual_file:  # noqa
                generated_actual_file.write(f'{stage},\n')
                with open(file_path.format(target='generated'), 'a+') as generated_file:  # noqa
                    generated_file.write(f'{stage},\n')

                    for module_path, existing_factories in factory_classes.items():  # noqa
                        if project not in module_path:
                            continue

                        sorted_existing_factories = (
                            sorted(
                                existing_factories,
                                key=lambda item: item.factory_class_name,
                            )
                        )

                        for existing_factory in sorted_existing_factories:
                            model = existing_factory.factory_class._meta.model
                            factory_class_name = (
                                existing_factory.factory_class_name
                            )

                            model.objects = model._base_manager

                            file = None

                            if (
                                model.objects.exists() and
                                factory_class_name.startswith(GENERATED_PREFIX)
                            ):
                                file = etalon_file
                            elif factory_class_name.startswith(GENERATED_ACTUAL_PREFIX):  # noqa
                                file = generated_actual_file
                            elif factory_class_name.startswith(GENERATED_PREFIX):  # noqa
                                file = generated_file

                            if file:
                                file.write(
                                    f'{module_path},{factory_class_name}\n'
                                )



import inspect
import json
import sys
import uuid
import warnings
from abc import (
    ABCMeta,
    abstractmethod,
)
from copy import (
    copy,
)
from importlib import (
    import_module,
)
from importlib.util import (
    module_from_spec,
    spec_from_file_location,
)
from pathlib import (
    Path,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from django.apps import (
    apps,
)
from django.conf import (
    settings,
)
from django.contrib.contenttypes.models import (
    ContentType,
)
from django.core.exceptions import (
    ValidationError,
)
from django.db.models import (
    Field,
    FileField,
    ForeignKey,
    Model,
    UUIDField,
)

from factory_bo.base import (
    DefaultManagerFactory,
)
from factory_bo.consts import (
    CONTENT_TYPE_MODEL,
    EXCLUDED_FIELD_CLASSES,
    FACTORY_CLASS_NAME_DELIMETER,
    GENERATED_PREFIX,
)
from factory_bo.enums import (
    FactoryUseTypeEnum,
    PreparingModelRecordTagEnum,
)
from factory_bo.factories import (
    ExistingFactory,
)
from factory_bo.helpers import (
    colored_stdout_output,
    substitute_model_get_queryset_method,
)
from factory_bo.records import (
    ModelRecord,
    PreparingModelRecord,
)
from factory_bo.signals import (
    existing_factory_prepare_filter,
)
from factory_bo.strings import (
    DISALLOWED_DJANGO_GET_OR_CREATE_FIELDS_FOUND_ERROR,
    FACTORIES_FOR_MODEL_NOT_FOUND,
    FACTORY_CLASS_FOR_MODEL_NOT_FOUND,
    MODEL_WITHOUT_DEFAULT_FACTORY_ERROR,
    RECORD_FOR_FK_FIELD_NOT_FOUND_ERROR,
    SEVERAL_MODEL_DAFAULT_FACTORIES_ERROR,
    USING_GENERATED_FACTORY_WARNING,
    WRONG_FACTORY_USE_TYPE_ERROR,
)


class ModelStorage:
    """
    Хранилище доступных моделей согласно подключенных плагинов, исключенных
    моделей и приложений

    Лучше использовать в связке с ExistingFactoryStorage, т.к. производится
    дополнительная проверка корректности фабрик
    """
    def __init__(self):
        self._models: Dict[str, Dict[str, Union[Type[Model], Dict[str, Field], Dict[str, str]]]] = {}  # noqa

        self._prepare()

    def _prepare_allowed_fields(
        self,
        model: Type[Model],
    ):
        """
        Добавление допустимых полей модели
        """
        allowed_fields = self._models[model._meta.label]['allowed_fields']

        for field in model._meta.fields:
            if (
                getattr(field, 'auto_now', False) or
                getattr(field, 'auto_now_add', False) or
                isinstance(field, FileField)
            ):
                continue

            allowed_fields[field.attname] = field

    def _prepare_foreign_key_fields_map(
        self,
        model: Type[Model],
    ):
        """
        Добавление полей внешних ключей с указанием моделей, на которые
        производится ссылка
        """
        foreign_key_fields_map = (
            self._models[model._meta.label]['foreign_key_fields_map']
        )
        processed_field_names = set()

        allowed_fields = self._models[model._meta.label]['allowed_fields']

        # Заполнение явных внешних ключей и
        # обработка ForeignKey
        fk_field_model_label_map = {
            field.attname: field.related_model._meta.label
            for field in allowed_fields.values() if (
                isinstance(field, ForeignKey) and
                field.attname not in processed_field_names
            )
        }

        foreign_key_fields_map.update(**fk_field_model_label_map)

        processed_field_names.update(fk_field_model_label_map.keys())

        # Заполнение псевдо внешних ключей указывающих на другие модели
        pseudo_fk_field_names = (
            set(settings.FACTORY_BO__PSEUDO_FK_IDS).intersection(
                allowed_fields.keys()
            ).difference(processed_field_names)
        )

        pseudo_fk_field_model_label_map = {
            field_name: settings.FACTORY_BO__PSEUDO_FK_IDS[field_name]
            for field_name in pseudo_fk_field_names
        }

        foreign_key_fields_map.update(**pseudo_fk_field_model_label_map)

        processed_field_names.update(pseudo_fk_field_names)

        # Заполнение псевдо внешних ключей, указывающих на ту же модель
        self_fk_field_names = (
            set(settings.FACTORY_BO__PSEUDO_SELF_FK_IDS).intersection(
                allowed_fields.keys()
            ).difference(processed_field_names)
        )

        self_fk_field_model_label_map = {
            field_name: model._meta.label
            for field_name in self_fk_field_names
        }

        foreign_key_fields_map.update(**self_fk_field_model_label_map)

    def _prepare(self):
        """
        Подготовка кеша моделей
        """
        allowed_models = filter(
            lambda model: not (
                model._meta.app_label in settings.FACTORY_BO__EXCLUDED_APPS or
                model._meta.label in settings.FACTORY_BO__EXCLUDED_MODELS or
                model._meta.proxy
            ),
            apps.get_models(include_auto_created=True)
        )

        for model in allowed_models:
            self._models[model._meta.label] = {
                'class_': model,
                'allowed_fields': {},
                'foreign_key_fields_map': {},
            }

            self._prepare_allowed_fields(
                model=model,
            )

            self._prepare_foreign_key_fields_map(
                model=model,
            )

    def get_model(
        self,
        model_label: str,
    ) -> Dict[str, Union[Type[Model], Dict[str, Field], Dict[str, str]]]:
        """
        Возвращает собранные данные модели
        """
        return self._models[model_label]

    def get_model_class(
        self,
        model_label: str,
    ) -> Optional[Type[Model]]:
        """
        Возращает класс модели из кеша по лейблу
        """
        model = None

        if model_label in self._models:
            model = self._models[model_label]['class_']

        return model

    def get_model_classes(self) -> List[Type[Model]]:
        """
        Возвращает все модели находящиеся в хранилище
        """
        return [
            model['class_']
            for model in self._models.values()
        ]

    def get_model_labels(self) -> List[str]:
        """
        Получение списка лейблов моделей, находящихся в хранилище
        """
        return [
            model['class_']._meta.label
            for model in self._models.values()
        ]

    def get_allowed_fields(
        self,
        model_label: str,
    ) -> Optional[List[Field]]:
        """
        Возращает допустимые поля модели
        """
        allowed_fields = None

        if model_label in self._models:
            allowed_fields = (
                self._models[model_label]['allowed_fields'].values()
            )

        return allowed_fields

    def get_allowed_field_names(
        self,
        model_label: str,
    ) -> Optional[List[str]]:
        """
        Возвращает имена допустимых полей
        """
        allowed_field_names = None

        if model_label in self._models:
            allowed_field_names = (
                self._models[model_label]['allowed_fields'].keys()
            )

        return allowed_field_names

    def get_model_foreign_key_fields_map(
        self,
        model_label: str,
    ):
        """
        Получение карты соответствия внешних ключей и лейблов моделей, на
        которые совершается ссылка
        """
        return self._models[model_label]['foreign_key_fields_map']

    def exists(
        self,
        model_label: str,
    ) -> bool:
        """
        Проверяет существование модели в хранилище
        """
        return model_label in self._models


class ExistingFactoryStorage:
    """
    Кеш существующих фабрик в проекте

    Кеш представляет из себя словарь, ключом которого является название фабрики,
    значением - именованный кортеж ExistingFactory с классом фабрики (
    factory_class) и путем модуля для импорта (import_path)
    """

    def __init__(
        self,
        model_storage: Optional[ModelStorage] = None,
        raise_on_nonexistent: Optional[bool] = True,
    ):
        if not model_storage:
            model_storage = ModelStorage()

        self._model_storage = model_storage
        self._model_factories_map: Dict[str, Dict[str, Union[ExistingFactory, Dict[str, ExistingFactory]]]] = {}  # noqa
        self._raise_on_nonexistent = raise_on_nonexistent

        self._prepare()

    @property
    def model_factories_map(self) -> Dict[str, Dict[str, Union[ExistingFactory, Dict[str, ExistingFactory]]]]:  # noqa
        return self._model_factories_map

    @property
    def model_storage(self) -> ModelStorage:
        return self._model_storage

    def get_model_factories_map(
        self,
        model_label: str,
    ) -> Dict[str, Union[ExistingFactory, Dict[str, ExistingFactory]]]:
        """
        Получение карты фабрик для модели
        """
        factories_map = self._model_factories_map.get(model_label)

        if not factories_map:
            raise ValueError(
                FACTORIES_FOR_MODEL_NOT_FOUND.format(
                    model_label=model_label,
                )
            )

        return factories_map

    def get_model_factories(
        self,
        model_label: str,
    ) -> Dict[str, ExistingFactory]:
        """
        Получение фабрик модели
        """
        factories_map = self.get_model_factories_map(
            model_label=model_label,
        )
        model_factories = {}

        model_factories.update(
            factories_map[FactoryUseTypeEnum.CUSTOM]
        )

        default_factory = factories_map[FactoryUseTypeEnum.DEFAULT]

        model_factories[default_factory.factory_class_name] = default_factory

        return model_factories

    def get_factory_class_for_model(
        self,
        model_label: str,
        factory_class_name: str,
    ) -> ExistingFactory:
        """
        Получение класса фабрики для модели
        """
        model_factories = self.get_model_factories(
            model_label=model_label,
        )

        if factory_class_name not in model_factories:
            raise ValueError(
                FACTORY_CLASS_FOR_MODEL_NOT_FOUND.format(
                    factory_class_name=factory_class_name,
                    model_label=model_label,
                )
            )

        if factory_class_name.startswith(GENERATED_PREFIX):
            colored_stdout_output(
                message=USING_GENERATED_FACTORY_WARNING.format(
                    generated_factory=factory_class_name,
                ),
                color=31,
            )

        return model_factories[factory_class_name]

    def get_model_default_factory(
        self,
        model_label: str,
    ) -> ExistingFactory:
        """
        Получение дефолтной фабрики модели
        """
        factories_map = self.get_model_factories_map(
            model_label=model_label,
        )

        return factories_map[FactoryUseTypeEnum.DEFAULT]

    def get_model_default_factory_map(self) -> Dict[str, ExistingFactory]:
        """
        Получение карты соответствия моделей и их дефолтных фабрик
        """
        model_default_factory_map = {}

        for model_label, factories_map in self._model_factories_map.items():
            default_factory = factories_map[FactoryUseTypeEnum.DEFAULT]

            model_default_factory_map[model_label] = default_factory

            if default_factory.factory_class_name.startswith(GENERATED_PREFIX):
                colored_stdout_output(
                    message=USING_GENERATED_FACTORY_WARNING.format(
                        generated_factory=default_factory.factory_class_name,
                    ),
                    color=31,
                )

        return model_default_factory_map

    def get_factories(self) -> List[ExistingFactory]:
        """
        Возвращает все фабрики хранилища
        """
        factories = []
        for factories_map in self._model_factories_map.values():
            factories.append(factories_map[FactoryUseTypeEnum.DEFAULT])

            factories.extend(factories_map[FactoryUseTypeEnum.CUSTOM].values())

        return factories

    def get_factory_class_by_name(
        self,
        factory_class_name: str,
    ) -> Optional[ExistingFactory]:
        """
        Метод получения класса фабрики по имени
        """
        existing_factory = None

        filtered_existing_factories = list(
            filter(
                lambda existing_factory: (
                    existing_factory and
                    existing_factory.factory_class_name == factory_class_name
                ),
                self.get_factories()
            )
        )

        if filtered_existing_factories:
            existing_factory = filtered_existing_factories[0]

        return existing_factory

    def _find_factories_modules(self) -> List[Tuple[str, str]]:
        """
        Поиск модулей содержащих фабрики
        """
        factories_modules = []

        checked_packed_paths = []
        for app_name in settings.INSTALLED_APPS:
            app_module = import_module(app_name)
            app_path = app_module.__path__

            # Если поиск уже осуществлялся по родительской директории,
            # то проверку нужно пропустить
            is_already_checked = False
            for cpp in checked_packed_paths:
                if app_path in cpp:
                    is_already_checked = True
                    break

            if is_already_checked:
                continue

            application_path = Path(app_path[0])
            factory_file_patterns = [
                '**/factories/**/*.py',
            ]

            for factory_file_pattern in factory_file_patterns:
                factory_modules_paths = application_path.glob(
                    factory_file_pattern
                )
                for factory_module_path in factory_modules_paths:
                    factory_module_path = str(factory_module_path)

                    # Нас не интересуют __init__.py
                    if '__init__.py' in factory_module_path:
                        continue

                    # Дополнительная фильтрация
                    filter_results = existing_factory_prepare_filter.send(
                        sender=app_name,
                        factory_module_path=factory_module_path,
                    )
                    filtered = False
                    for _, response in filter_results:
                        if not response:
                            filtered = True
                            break

                    if filtered:
                        continue

                    module_name = (
                        str(factory_module_path).split('/')[-1].split('.')[0]
                    )

                    spec = spec_from_file_location(
                        name=module_name,
                        location=factory_module_path,
                    )
                    factory_module = module_from_spec(
                        spec=spec,
                    )
                    spec.loader.exec_module(factory_module)
                    factories_modules.append(
                        (
                            factory_module,
                            factory_module_path,
                        )
                    )

                checked_packed_paths.append(app_path)

        return factories_modules

    def _get_module_import_path(self, module, sys_path) -> str:
        """
        Предназначен для получения пути пакета для импорта фабрики
        """
        module_path = module.__file__

        package_path = max(
            filter(
                lambda path: path in module_path,
                sys_path
            )
        )

        relative_module_path = module_path.split(f'{package_path}/')[1]
        import_path = '.'.join(relative_module_path.split('.')[0].split('/'))

        return import_path

    def _check_factory_class(
        self,
        factory_class: DefaultManagerFactory,
    ):
        """
        Проверка фабрики на допустимость к использованию и корректность
        """
        model_label = factory_class.get_model_label()

        is_allowed_factory = False

        if self._model_storage.exists(model_label):
            is_allowed_factory = True

        # TODO BOBUH-15824 После полного перехода на новые фабрики, в условии
        #  необходимо убрать проверку на тип использования - кастомный.
        #  Проверку должны проходить все фабрики без исключения
        if (
            is_allowed_factory and
            FactoryUseTypeEnum.is_default(factory_class.get_factory_use_type())
        ):
            wrong_fields = set(
                factory_class.get_django_get_or_create()
            ).difference(
                self._model_storage.get_allowed_field_names(model_label)
            )

            if wrong_fields:
                raise ValidationError(
                    DISALLOWED_DJANGO_GET_OR_CREATE_FIELDS_FOUND_ERROR.format(
                        disallowed_fields=', '.join(wrong_fields),
                        factory_class=factory_class.__name__,
                    )
                )

            wrong_excluded_fields = set(
                factory_class.get_django_get_or_create()
            ).intersection(
                factory_class.get_excluded_fk_fields()
            )

            if wrong_excluded_fields:
                raise ValidationError(
                    DISALLOWED_DJANGO_GET_OR_CREATE_FIELDS_FOUND_ERROR.format(
                        disallowed_fields=', '.join(wrong_excluded_fields),
                        factory_class=factory_class.__name__,
                    )
                )

            factory_use_type = factory_class.get_factory_use_type()

            if not FactoryUseTypeEnum.exists(factory_use_type):
                raise ValidationError(
                    WRONG_FACTORY_USE_TYPE_ERROR.format(
                        factory_use_type=factory_use_type,
                        factory_class_name=factory_class.__name__,
                    )
                )

            model_factories = (
                self._model_factories_map[model_label] if
                model_label in self._model_factories_map else
                None
            )

            if (
                model_factories and
                FactoryUseTypeEnum.is_default(factory_use_type) and
                model_factories[FactoryUseTypeEnum.DEFAULT] and
                factory_class.__name__ != model_factories[FactoryUseTypeEnum.DEFAULT].factory_class.__name__  # noqa
            ):
                second_factory_class_name = (
                    model_factories[FactoryUseTypeEnum.DEFAULT].factory_class.__name__# noqa
                )

                raise ValidationError(
                    SEVERAL_MODEL_DAFAULT_FACTORIES_ERROR.format(
                        model_label=model_label,
                        factories=(
                            f'{factory_class.__name__}, '
                            f'{second_factory_class_name}'
                        ),
                    )
                )

        return is_allowed_factory

    def _check_default_factory_model_exists(self):
        """
        Проверка существования дефолных фабрик для всех допустимых моделей.
        Если для модели нет дефолтной фабрики - нужно падать с ошибкой
        """
        model_default_factory_map = self.get_model_default_factory_map()

        models_without_default_factory = set()

        for model_label in self._model_storage.get_model_labels():
            if model_label not in model_default_factory_map:
                models_without_default_factory.add(model_label)

        if models_without_default_factory:
            error_message = MODEL_WITHOUT_DEFAULT_FACTORY_ERROR.format(
                model_labels=', '.join(sorted(models_without_default_factory)),
            )

            if self._raise_on_nonexistent:
                raise ValueError(error_message)
            else:
                warnings.warn(error_message)

    def _get_factory_class_name_alias(
        self,
        factory_class: DefaultManagerFactory,
    ) -> Optional[str]:
        """
        Возвращает сгенерированный алиас имени фабрики, если существует
        фабрика с таким же именем
        """
        factory_class_name_alias = None
        existing_factory_class = self.get_factory_class_by_name(
            factory_class_name=factory_class.__name__,
        )

        if (
            existing_factory_class and
            existing_factory_class.factory_model_label != factory_class.get_model_label()  # noqa
        ):
            uuid_hex_part = uuid.uuid4().hex[:6]

            factory_class_name_alias = (
                f'{factory_class.__name__}{FACTORY_CLASS_NAME_DELIMETER}'
                f'{uuid_hex_part}'
            )

        return factory_class_name_alias

    def _prepare(self):
        """
        Поиск фабрик во всех подключенных приложениях
        """
        sys_path = set(sys.path)

        factories_modules = self._find_factories_modules()
        for factory_module, module_path in factories_modules:
            factory_classes_names = list(filter(
                lambda name: (
                    inspect.isclass(getattr(factory_module, name)) and
                    issubclass(getattr(factory_module, name), DefaultManagerFactory) and  # noqa
                    not getattr(factory_module, name)._meta.abstract
                ),
                dir(factory_module)
            ))

            if factory_classes_names:
                import_path = self._get_module_import_path(
                    module=factory_module,
                    sys_path=sys_path,
                )

                for factory_class_name in factory_classes_names:
                    factory_class: DefaultManagerFactory = getattr(
                        factory_module,
                        factory_class_name
                    )

                    model_label = factory_class.get_model_label()

                    if not self._check_factory_class(factory_class):
                        continue

                    factory_class_name_alias = (
                        self._get_factory_class_name_alias(
                            factory_class=factory_class,
                        )
                    )

                    if model_label not in self._model_factories_map:
                        self._model_factories_map[model_label] = {
                            FactoryUseTypeEnum.DEFAULT: None,
                            FactoryUseTypeEnum.CUSTOM: {},
                        }

                    model_factories = self._model_factories_map[model_label]
                    factory_use_type = factory_class.get_factory_use_type()

                    existing_factory = ExistingFactory(
                        factory_class=factory_class,
                        factory_class_name_alias=factory_class_name_alias,
                        import_path=import_path,
                        module_path=module_path,
                    )

                    if FactoryUseTypeEnum.is_default(factory_use_type):
                        model_factories[FactoryUseTypeEnum.DEFAULT] = (
                            existing_factory
                        )
                    elif FactoryUseTypeEnum.is_custom(factory_use_type):
                        model_factories[FactoryUseTypeEnum.CUSTOM][factory_class_name] = (  # noqa
                            existing_factory
                        )

        self._check_default_factory_model_exists()


class PreparingModelRecordStorage:
    """
    Хранилище обрабатываемых записей моделей
    """

    def __init__(
        self,
        model_factory_correlation_storage: 'ModelFactoryCorrelationStorage',
        existing_factory_storage: ExistingFactoryStorage,
    ):
        self._records: Dict[Tuple[str, str], PreparingModelRecord] = {}

        self._model_factory_correlation_storage = (
            model_factory_correlation_storage
        )

        self._existing_factory_storage: ExistingFactoryStorage = (
            existing_factory_storage
        )

    @property
    def records(self) -> Dict[Tuple[str, str], PreparingModelRecord]:
        return self._records

    def _fill_fk_fields(
        self,
        model_record: ModelRecord,
        fk_source_storage: 'PreparingModelRecordStorage',
        fk_field_model_label_map: Dict[str, str],
    ):
        """
        Заполнение полей внешних ключей
        """
        for field_name, model_label in fk_field_model_label_map.items():
            record_field = model_record.fields.get(field_name)

            if record_field:
                if not isinstance(
                    record_field,
                    PreparingModelRecord,
                ):
                    field_value_id = model_record.fields.pop(field_name)

                    if field_value_id == 'None':
                        field_value = 'None'
                    else:
                        key = (
                            model_label,
                            field_value_id,
                        )

                        try:
                            field_value = (
                                fk_source_storage.records.get(key) or
                                self._records[key]
                            )
                        except KeyError:
                            raise KeyError(
                                RECORD_FOR_FK_FIELD_NOT_FOUND_ERROR.format(
                                    model_label=model_record.model_label,
                                    field_name=field_name,
                                    related_record=model_label,
                                    field_value_id=field_value_id,
                                )
                            )

                    model_record.fields[field_name] = field_value

            else:
                # нужно проверить, можно ли подставлять пустые значения
                model_field = model_record.model_class._meta.get_field(field_name)  # noqa

                if model_field.null:
                    field_value = 'None'

                elif model_field.blank:
                    field_value = '\'\''

                else:
                    raise ValueError(
                        f'A value for field {field_name} in model {model_label}'
                        f' must be provided in the fixture!'
                    )

                model_record.fields[field_name] = field_value

    def get_record(
        self,
        model_label: str,
        record_pk: str,
    ):
        """
        Метод получения записи их хранилища
        """
        key = (
            model_label,
            record_pk,
        )

        return self._records[key]

    def get_dependencies_for_record_by_fk_recursively(
        self,
        preparing_model_record: PreparingModelRecord,
        result: Optional[List[PreparingModelRecord]] = None,
    ) -> List[PreparingModelRecord]:
        """Формирует список зависимостей для записи модели на основе fk-полей (рекурсивно).

        Args:
            preparing_model_record: Запись модели, для которой нужно сформировать список зависимостей,
            result: Рекурсивно наполняемый список.

        Returns:
            Список записей моделей.

        """
        if result is None:
            result = []
        else:
            result.append(preparing_model_record)

        model_record: ModelRecord = preparing_model_record.model_record
        existing_factory: ExistingFactory = self._model_factory_correlation_storage.get_existing_factory(
            model_record.model_label
        )

        if model_record.fk_field_names:
            for field_name in model_record.fk_field_names:
                if (
                    model_record.fields[field_name] == 'None' or
                    field_name not in existing_factory.get_or_create_fields
                ):
                    continue

                if isinstance(model_record.fields[field_name], str):
                    foreign_preparing_model_record = self.get_record(
                        model_record.foreign_key_fields_map[field_name],
                        model_record.fields[field_name]
                    )
                else:
                    foreign_preparing_model_record = model_record.fields[field_name]

                self.get_dependencies_for_record_by_fk_recursively(
                    foreign_preparing_model_record,
                    result=result,
                )

        return result

    def add(
        self,
        model_record: ModelRecord,
        tag: int,
        is_add_recursively_fks: bool = False,
        use_factory_get_or_create_fields: bool = False,
        changed_fields: Optional[Dict[str, Any]] = None,
        is_update_model_record: bool = False,
    ) -> PreparingModelRecord:
        """
        Добавление/обновление обрабатываемой записи в хранилище. При необходимости можно
        указать, чтобы в хранилище были добавлены все ссылки на внешние объекты
        в хранилище.
        """
        key = model_record.key

        if key in self._records:
            preparing_model_record = self._records[key]

            if is_update_model_record:
                preparing_model_record.model_record = model_record
            else:
                preparing_model_record.set_tag(
                    tag=tag,
                    changed_fields=changed_fields,
                )
        else:
            preparing_model_record = PreparingModelRecord(
                model_record=model_record,
                tag=tag,
                changed_fields=changed_fields,
            )

            self._records[key] = preparing_model_record

            if is_add_recursively_fks:
                model_default_factory = self._existing_factory_storage.get_model_default_factory(
                    model_record.model_label
                )

                for field_name in model_record.fk_field_names:
                    if (
                        use_factory_get_or_create_fields and
                        field_name not in model_default_factory.get_or_create_fields
                    ):
                        continue

                    field_value: Union[PreparingModelRecord, str] = (
                        model_record.fields[field_name]
                    )

                    if field_value != 'None':
                        self.add(
                            model_record=field_value.model_record,
                            tag=PreparingModelRecordTagEnum.GETTING,
                            is_add_recursively_fks=True,
                            use_factory_get_or_create_fields=use_factory_get_or_create_fields,
                        )

        return preparing_model_record

    def set_dependencies_recursively(
        self,
        model_record: ModelRecord,
        has_dependence: bool = False,
    ) -> PreparingModelRecord:
        """
        Рекурсивное проставление зависимостей для записей в хранилище,
        с учетом зависимости предыдущей(родительской) записи.
        """
        preparing_record = self._records[model_record.key]
        correlation_storage = self._model_factory_correlation_storage

        if not preparing_record.has_dependence:
            preparing_record.has_dependence = has_dependence
            factory = correlation_storage.get_existing_factory(
                model_label=model_record.model_label,
            )

            for field_name in model_record.fk_field_names:
                preparing_record_from_fk: Union[PreparingModelRecord, str] = (
                    model_record.fields[field_name]
                )

                if preparing_record.has_only_getting_tag:
                    has_dependence = (
                            field_name in factory.get_or_create_fields and
                            preparing_record.has_dependence
                    )
                elif preparing_record.has_creating_tag:
                    has_dependence = True

                elif preparing_record.has_deleting_tag:
                    has_dependence = field_name in factory.get_or_create_fields

                elif preparing_record.has_updating_tag:
                    has_dependence = field_name in factory.get_or_create_fields

                    if field_name in preparing_record.changed_fields:
                        changed_field_value: Union[PreparingModelRecord, str] = (
                            preparing_record.changed_fields[field_name]
                        )
                        self.set_dependencies_recursively(
                            model_record=changed_field_value.model_record,
                            has_dependence=True,
                        )

                if preparing_record_from_fk != 'None':
                    if model_record.foreign_key_fields_map[field_name] == (
                        CONTENT_TYPE_MODEL
                    ):
                        continue

                    self.set_dependencies_recursively(
                        model_record=preparing_record_from_fk.model_record,
                        has_dependence=has_dependence,
                    )

        return preparing_record

    def substitute_fk_ids(
        self,
        fk_source_storage: Optional['PreparingModelRecordStorage'] = None,
    ):
        """
        Замена полей внешних ключей с идентификаторов на объекты. Убирается
        суффикс _id и устанавливается значение. Если значение не найдено, то
        устанавливается 'None'.

        В параметрах можно передать fk_source_storage - хранилище, из которого
        должны проставляться значения внешних ключей. Если хранилище не
        передано, то значения будут браться из текущего хранилища. Это сделано
        для работы нахождения разницы состояний базы данных.

        Сама замена производится после наполнения хранилища записями
        """
        if not fk_source_storage:
            fk_source_storage = self

        for preparing_model_record in self._records.values():
            model_record = preparing_model_record.model_record

            self._fill_fk_fields(
                model_record=model_record,
                fk_source_storage=fk_source_storage,
                fk_field_model_label_map=model_record.foreign_key_fields_map,
            )

    def difference(
        self,
        storage: 'PreparingModelRecordStorage',
    ) -> 'PreparingModelRecordStorage':
        """
        Получает разницу двух хранилищ в виде нового хранилища со всеми
        зависимыми записями моделей
        """
        from factory_bo.actions import (
            DifferencePreparingModelRecordStorageAction,
        )

        difference_action = DifferencePreparingModelRecordStorageAction(
            begin_storage=storage,
            end_storage=self,
            model_factory_correlation_storage=(
                self._model_factory_correlation_storage
            ),
            existing_factory_storage=self._existing_factory_storage,
        )

        difference_storage = difference_action.run()

        return difference_storage

    def union(
        self,
        src_storage: 'PreparingModelRecordStorage',
    ) -> 'PreparingModelRecordStorage':
        """
        Объединение двух хранилищ. Считается, что хранилище, у которого вызван
        метод - является целевым и результирующее хранилище будет создано на
        его основе. В метод передается хранилище-донор, записи которого будут
        перемещены в результирующее хранилище.
        """
        from factory_bo.actions import (
            UnionPreparingModelRecordStorageAction,
        )

        union_action = UnionPreparingModelRecordStorageAction(
            dst_storage=self,
            src_storage=src_storage,
            model_factory_correlation_storage=(
                self._model_factory_correlation_storage
            ),
            existing_factory_storage=self._existing_factory_storage,
        )

        union_storage = union_action.run()

        return union_storage


class PreparingModelRecordStorageCreator(metaclass=ABCMeta):
    """
    Базовый класс для создания хранилищ кешей записей фабрик
    """

    def __init__(
        self,
        existing_factory_storage: ExistingFactoryStorage,
        *args,
        **kwargs,
    ):
        self._existing_factory_storage = existing_factory_storage
        self._content_type_cache: Dict[int, ContentType] = {
            ct.pk: ct
            for ct in ContentType.objects.all()
        }

    @abstractmethod
    def create(
        self,
        need_substitute_fk_ids: bool = False,
    ) -> PreparingModelRecordStorage:
        """
        Контрактный метод создания хранилища кешей фабрик
        """
        pass


class DefaultDBPreparingModelRecordStorageCreator(
    PreparingModelRecordStorageCreator,
):
    """
    Создатель хранилища кешей записей фабрик из дефолтной базы данных
    """

    def create(
        self,
        need_substitute_fk_ids: bool = False,
        fk_source_storage: Optional['PreparingModelRecordStorage'] = None,
        preparing_model_record_tag: int = PreparingModelRecordTagEnum.CREATING,
    ) -> PreparingModelRecordStorage:
        """
        Создание хранилища кешей фабрик и их наполнение
        """
        correlation_storage_creator = ModelFactoryCorrelationStorageCreator(
            existing_factory_storage=self._existing_factory_storage,
        )

        correlation_storage = correlation_storage_creator.create()

        preparing_model_record_storage = PreparingModelRecordStorage(
            model_factory_correlation_storage=correlation_storage,
            existing_factory_storage=self._existing_factory_storage,
        )
        model_storage = self._existing_factory_storage.model_storage

        for model in model_storage.get_model_classes():
            objects = self.get_model_objects(model)

            if objects:
                self.add_model_objects_to_storage(
                    model,
                    model_storage,
                    objects,
                    preparing_model_record_storage,
                    preparing_model_record_tag,
                )

        if need_substitute_fk_ids:
            preparing_model_record_storage.substitute_fk_ids(
                fk_source_storage=fk_source_storage,
            )

        return preparing_model_record_storage

    def update_storage(
        self,
        preparing_model_record_storage: Optional['PreparingModelRecordStorage'],
        changed_tables: set,
        need_substitute_fk_ids: bool = False,
        preparing_model_record_tag: int = PreparingModelRecordTagEnum.CREATING,
    ) -> PreparingModelRecordStorage:
        """
        Обновление хранилища кешей фабрик по списку таблиц.
        """
        model_storage = self._existing_factory_storage.model_storage

        for model in model_storage.get_model_classes():
            if changed_tables and model._meta.db_table not in changed_tables:
                continue

            objects = self.get_model_objects(model)

            if objects:
                self.add_model_objects_to_storage(
                    model,
                    model_storage,
                    objects,
                    preparing_model_record_storage,
                    preparing_model_record_tag,
                    is_update_model_record=True,
                )

        if need_substitute_fk_ids:
            preparing_model_record_storage.substitute_fk_ids()

        return preparing_model_record_storage

    def get_model_objects(
        self,
        model: Type[Model],
    ) -> list:
        """
        Получить все объекты модели используя базовый менеджер.
        """
        with substitute_model_get_queryset_method(model):
            try:
                objects = list(model.objects.order_by('pk'))
            except Exception as e:
                objects = []
                print(f'model "{model.__name__}" failed due to: {e}')

        return objects

    def add_model_objects_to_storage(
        self,
        model: Type[Model],
        model_storage: ModelStorage,
        objects: list,
        preparing_model_record_storage: PreparingModelRecordStorage,
        preparing_model_record_tag: int,
        is_update_model_record: bool = False,
    ):
        """
        Добавить запись объекта модели в хранилище и удалить из хранилища лишние записи.

        Args:
            model: Django-модель
            model_storage: Хранилище моделей
            objects: QuerySet-объектов модели
            preparing_model_record_storage: Хранилище записей моделей
            preparing_model_record_tag: Тег для указания в preparing_model_record
            is_update_model_record: Ключ требуется ли обновить запись model_record
        """
        exists_model_record_keys = {
            key
            for key in preparing_model_record_storage.records.keys()
            if key[0] == model._meta.label
        }
        model_fk_fields_map = model_storage.get_model_foreign_key_fields_map(
            model_label=model._meta.label,
        )
        allowed_fields = model_storage.get_allowed_fields(
            model_label=model._meta.label,
        )

        for object_ in objects:
            model_record = self.prepare_model_record(
                model,
                model_fk_fields_map,
                allowed_fields,
                object_,
            )
            if model_record.key in exists_model_record_keys:
                exists_model_record_keys.remove(model_record.key)

            preparing_model_record_storage.add(
                model_record=model_record,
                tag=preparing_model_record_tag,
                is_update_model_record=is_update_model_record,
            )

        if exists_model_record_keys:
            # Удалим значения по оставшимся ключам, т.к. не были найдены соответствующие объекты модели.
            for key in exists_model_record_keys:
                preparing_model_record_storage.records.pop(key)

    def prepare_model_record(
        self,
        model: Type[Model],
        model_fk_fields_map: Dict[str, str],
        allowed_fields: Optional[List[Field]],
        object_
    ) -> ModelRecord:
        """
        Подготовка Записи модели
        """
        fields = {}

        for field in allowed_fields:
            if field.__class__.__name__ in EXCLUDED_FIELD_CLASSES:
                fields[field.attname] = 'None'
            else:
                field_value = getattr(object_, field.attname)

                if (
                    isinstance(field, UUIDField) and
                    field_value is not None
                ):
                    field_value = repr(str(field_value))
                elif (field.primary_key or isinstance(field, ForeignKey)) and isinstance(field_value, str):
                    # строковые значения PK и FK не преобразовываем
                    pass
                else:
                    field_value = repr(field_value)

                fields[field.attname] = field_value

        model_record = ModelRecord(
            model_class=model,
            foreign_key_fields_map=model_fk_fields_map,
            fields=fields,
            content_type_cache=self._content_type_cache,
        )
        return model_record


class JSONPreparingModelStorageCreator(PreparingModelRecordStorageCreator):
    """
    Создатель хранилища кешей записей тестовых фабрик из переданного контента
    json-файла
    """

    def __init__(
        self,
        factories_json: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self._models_dict = json.loads(factories_json)

    def create(
        self,
        need_substitute_fk_ids: bool = False,
        fk_source_storage: Optional['PreparingModelRecordStorage'] = None,
        preparing_model_record_tag: int = PreparingModelRecordTagEnum.CREATING,
    ) -> PreparingModelRecordStorage:
        """
        Создание хранилища кешей с заполнением записей из переданного контента
        """
        correlation_storage_creator = ModelFactoryCorrelationStorageCreator(
            existing_factory_storage=self._existing_factory_storage,
        )

        correlation_storage = correlation_storage_creator.create()

        preparing_model_record_storage = PreparingModelRecordStorage(
            model_factory_correlation_storage=correlation_storage,
            existing_factory_storage=self._existing_factory_storage,
        )

        for model_label, models_list in self._models_dict.items():
            model_class = self._existing_factory_storage.model_storage.get_model_class(  # noqa
                model_label=model_label,
            )

            # Модели не находящиеся в хранилище не соответствуют предъявляемым
            # требованиям
            if not model_class:
                continue

            for model_fields in models_list:
                model_record = ModelRecord(
                    model_class=model_class,
                    foreign_key_fields_map=self._existing_factory_storage.model_storage.get_model_foreign_key_fields_map(  # noqa
                        model_label=model_label,
                    ),
                    fields=model_fields,
                    content_type_cache=self._content_type_cache,
                )

                preparing_model_record_storage.add(
                    model_record=model_record,
                    tag=preparing_model_record_tag,
                )

        if need_substitute_fk_ids:
            preparing_model_record_storage.substitute_fk_ids(
                fk_source_storage=fk_source_storage,
            )

        return preparing_model_record_storage


class ModelFactoryCorrelationStorage:
    """
    Хранилище соответствий моделей и фабрик
    """

    def __init__(
        self,
        correlation_map: Optional[Dict[str, ExistingFactory]],
    ):
        self._correlation_map = correlation_map or {}

    def get_existing_factory(
        self,
        model_label: str,
    ):
        """
        Метод получения фабрики для модели
        """
        return self._correlation_map[model_label]


class ModelFactoryCorrelationStorageCreator:
    """
    Создатель хранилищ соответствий моделей и фабрик. Упрощает работу с
    созданием хранилища, т.к. заблаговременно создается дефолтная карта
    соответствия моделей и фабрик.
    """

    def __init__(
        self,
        existing_factory_storage: ExistingFactoryStorage,
    ):
        self._existing_factory_storage = existing_factory_storage

        self._default_correlation_map = (
            self._existing_factory_storage.get_model_default_factory_map()
        )

    def create(
        self,
        custom_correlation_map: Optional[Dict[str, str]] = None,
    ):
        """
        Фабричный метод создания хранилища соответствия моделей и фабрик с
        поправкой на кастомные значения
        """
        correlation_map = copy(self._default_correlation_map)

        if custom_correlation_map:
            for model_label, factory_class_name in custom_correlation_map.items():  # noqa
                factory = self._existing_factory_storage.get_factory_class_for_model(  # noqa
                    model_label=model_label,
                    factory_class_name=factory_class_name,
                )

                correlation_map[model_label] = factory

        return ModelFactoryCorrelationStorage(
            correlation_map=correlation_map,
        )



from collections import (
    Iterable,
)
from copy import (
    deepcopy,
)
from random import (
    randint,
)
from typing import (
    Dict,
    Optional,
    Tuple,
)

from django.core.exceptions import (
    ValidationError,
)

from factory_bo.consts import (
    MODEL_CLASS_NAME_ID_DELIMITER,
    PK_OFFSET,
)
from factory_bo.enums import (
    PreparingModelRecordTagEnum,
)
from factory_bo.factories import (
    ExistingFactory,
)
from factory_bo.mixins import (
    PreparingModelSortMixin,
)
from factory_bo.records import (
    PreparingModelRecord,
)
from factory_bo.storages import (
    ExistingFactoryStorage,
    ModelFactoryCorrelationStorage,
    ModelFactoryCorrelationStorageCreator,
    PreparingModelRecordStorage,
)
from factory_bo.strings import (
    NON_UNIQUE_DJANGO_GET_OR_CREATE_RECORD_FOUND_ERROR,
)


class DifferencePreparingModelRecordStorageAction(PreparingModelSortMixin):
    """
    Служит для нахождения разницы состояний записей хранилищ. Результатом
    работы является новое хранилище, содержащее разницу между двумя хранилищами
    """
    def __init__(
        self,
        begin_storage: PreparingModelRecordStorage,
        end_storage: PreparingModelRecordStorage,
        model_factory_correlation_storage: ModelFactoryCorrelationStorage,
        existing_factory_storage: ExistingFactoryStorage,
    ):
        self._begin_storage = begin_storage
        self._end_storage = end_storage
        self._model_factory_correlation_storage = (
            model_factory_correlation_storage
        )
        self._existing_factory_storage = existing_factory_storage

    def _find_creating_difference(
        self,
        difference_storage: PreparingModelRecordStorage,
        begin_storage: PreparingModelRecordStorage,
        end_storage: PreparingModelRecordStorage,
    ):
        """
        Поиск и добавление создаваемых записей
        """
        different_record_keys = set(end_storage.records.keys()).difference(
            begin_storage.records.keys()
        )

        for key in different_record_keys:
            preparing_model_record = end_storage.records[key]

            difference_storage.add(
                model_record=preparing_model_record.model_record,
                tag=PreparingModelRecordTagEnum.CREATING,
                is_add_recursively_fks=True,
            )

    def _find_updating_difference(
        self,
        difference_storage: PreparingModelRecordStorage,
        begin_storage: PreparingModelRecordStorage,
        end_storage: PreparingModelRecordStorage,
    ):
        """
        Поиск и добавление обновляемых записей
        """
        intersection_record_keys = set(end_storage.records.keys()).intersection(
            begin_storage.records.keys()
        )

        for key in intersection_record_keys:
            begin_preparing_model_record = begin_storage.records.get(key)
            end_preparing_model_record = end_storage.records.get(key)

            if begin_preparing_model_record and end_preparing_model_record:
                changed_fields = {}

                field_names = begin_preparing_model_record.model_record.fields.keys()  # noqa

                for field_name in field_names:
                    begin_field_value = begin_preparing_model_record.model_record.fields[field_name]  # noqa
                    end_field_value = end_preparing_model_record.model_record.fields[field_name]  # noqa

                    if begin_field_value != end_field_value:
                        changed_fields[field_name] = end_field_value

                        if isinstance(end_field_value, PreparingModelRecord):
                            difference_storage.add(
                                model_record=end_field_value.model_record,
                                tag=PreparingModelRecordTagEnum.GETTING,
                                is_add_recursively_fks=True,
                            )

                if changed_fields:
                    difference_storage.add(
                        model_record=begin_preparing_model_record.model_record,
                        tag=PreparingModelRecordTagEnum.UPDATING,
                        is_add_recursively_fks=True,
                        changed_fields=changed_fields,
                    )

    def _find_deleting_difference(
        self,
        difference_storage: PreparingModelRecordStorage,
        begin_storage: PreparingModelRecordStorage,
        end_storage: PreparingModelRecordStorage,
    ):
        """
        Поиск и обновление удаляемых записей
        """
        different_record_keys = set(begin_storage.records.keys()).difference(
            end_storage.records.keys()
        )

        for key in different_record_keys:
            preparing_model_record = begin_storage.records[key]

            difference_storage.add(
                model_record=preparing_model_record.model_record,
                tag=PreparingModelRecordTagEnum.DELETING,
                is_add_recursively_fks=True,
            )

    def _set_dependencies(
        self,
        difference_storage: PreparingModelRecordStorage,
    ):
        """
        Указание зависимостей в записях
        """
        for preparing_model_record in difference_storage.records.values():
            difference_storage.set_dependencies_recursively(
                model_record=preparing_model_record.model_record,
            )

    def run(self) -> PreparingModelRecordStorage:
        """
        Получает разницу двух хранилищ в виде нового хранилища со всеми
        зависимыми записями моделей
        """
        correlation_storage_creator = ModelFactoryCorrelationStorageCreator(
            existing_factory_storage=self._existing_factory_storage,
        )

        correlation_storage = correlation_storage_creator.create()

        difference_storage = PreparingModelRecordStorage(
            model_factory_correlation_storage=correlation_storage,
            existing_factory_storage=self._existing_factory_storage,
        )

        self._find_creating_difference(
            difference_storage=difference_storage,
            begin_storage=self._begin_storage,
            end_storage=self._end_storage,
        )

        self._find_updating_difference(
            difference_storage=difference_storage,
            begin_storage=self._begin_storage,
            end_storage=self._end_storage,
        )

        self._find_deleting_difference(
            difference_storage=difference_storage,
            begin_storage=self._begin_storage,
            end_storage=self._end_storage,
        )

        self._set_dependencies(
            difference_storage=difference_storage,
        )

        return difference_storage


class UnionPreparingModelRecordStorageAction(PreparingModelSortMixin):
    """
    Действие объединения двух хранилищ - целевого (dst_storage) и донора (
    src_storage), с подготовленными записями моделей

    Объединение хранилищ производится по следующему алгоритму:

    1) Создается полная копия целевого хранилища и сохраняется в качестве
        результирующего;
    2) Производится сортировка записей хранилища-донора по возрастанию степени
        зависимости записей;
    3) Производится добавление записей хранилище-донор. При переносе записи
        производится подмена идентификаторов записей во внешних ключей из
        таблицы соответствия. В качестве внешнего хранилища указывается
        результирующее хранилище;
    4) В результате должено быть получено объединенное результирующее хранилище.
    """
    def __init__(
        self,
        dst_storage: PreparingModelRecordStorage,
        src_storage: PreparingModelRecordStorage,
        model_factory_correlation_storage: ModelFactoryCorrelationStorage,
        existing_factory_storage: ExistingFactoryStorage,
    ):
        self._dst_storage = dst_storage
        self._src_storage = src_storage
        self._result_storage: Optional[PreparingModelRecordStorage] = None
        self._model_factory_correlation_storage = (
            model_factory_correlation_storage
        )
        self._existing_factory_storage = existing_factory_storage
        self._model_default_factory_map: Dict[str, ExistingFactory] = (
            self._existing_factory_storage.get_model_default_factory_map()
        )

        self._table_conformity: Dict[Tuple[str, str], PreparingModelRecord] = {}
        self._preparing_result_model_records_map: Dict[Tuple[str, Tuple[str, ...]], Tuple[str, str]] = {}  # noqa

    def _prepare_result_storage(self):
        """
        Создание результирующего хранилища
        """
        self._result_storage = deepcopy(self._dst_storage)

        # Все записи нужно пометить для выборки
        for record in self._result_storage.records.values():
            self._result_storage.add(
                model_record=record.model_record,
                tag=PreparingModelRecordTagEnum.GETTING,
            )

    def _get_unique_preparing_model_record_key(
        self,
        record: PreparingModelRecord,
    ):
        """
        Рекурсивное получение уникального ключа обрабатываемой записи модели.
        Если в наборе полей django_get_or_create фабрики присутствуют внешние
        ключи, и в качестве значений указаны экземпляры PreparingModelRecord, то
        необходимо для них получить уникальный ключ
        """
        existing_factory = (
            self._model_default_factory_map[record.model_record.model_label]
        )
        get_or_create_key_getter = (
            existing_factory.factory_class.get_django_get_or_create_getter()
        )
        record_get_or_create_key = []

        key_items = get_or_create_key_getter(record.model_record.all_fields)

        if not isinstance(key_items, Iterable) or isinstance(key_items, str):
            key_items = [
                key_items,
            ]

        for key_item in key_items:
            if isinstance(key_item, PreparingModelRecord):
                key_item = self._get_unique_preparing_model_record_key(
                    record=key_item,
                )

            record_get_or_create_key.append(key_item)

        return tuple(record_get_or_create_key)

    def _prepare_preparing_result_model_records_map(self):
        """
        Подготовка карты соответствия набора значений записей полей
        django_get_or_create и идентификатора
        """
        for (model_label, record_id), record in self._result_storage.records.items():  # noqa
            existing_factory = self._model_default_factory_map[model_label]
            model_get_or_create_key = (
                model_label,
                self._get_unique_preparing_model_record_key(
                    record=record,
                ),
            )

            if model_get_or_create_key in self._preparing_result_model_records_map:  # noqa
                raise ValidationError(
                    NON_UNIQUE_DJANGO_GET_OR_CREATE_RECORD_FOUND_ERROR.format(
                        factory_class_name=existing_factory.factory_class_name,
                        key=model_get_or_create_key,
                    )
                )

            self._preparing_result_model_records_map[model_get_or_create_key] = (  # noqa
                model_label,
                record_id,
            )

    def _recalculate_src_storage_record_pks(self):
        """
        Перед объединением двух хранилищ, необходимо произвести сдвиг всех
        идентификаторов на константное значение для избежания коллизий при
        появлении разных записей с одинаковыми идентификаторами
        """
        replace_keys_map = {}

        src_records = self._src_storage.records

        for (model_label, pk), record in src_records.items():
            if record.model_record.pk.isdigit():
                while True:
                    new_pk = str(int(record.model_record.pk) + randint(100_000, PK_OFFSET))
                    new_key = (
                        model_label,
                        new_pk,
                    )

                    if new_key not in src_records:
                        break

                record.model_record.pk = new_pk

                replace_keys_map[(model_label, pk)] = new_key

        for old_key, new_key in replace_keys_map.items():
            src_records[new_key] = src_records.pop(old_key)

    def _transfer_src_record(
        self,
        src_record_key: str,
    ):
        """
        Перенос записи из хранилища-донора в результирующее хранилище
        """
        model_label, record_pk = src_record_key.split(
            MODEL_CLASS_NAME_ID_DELIMITER
        )

        src_record = self._src_storage.get_record(
            model_label=model_label,
            record_pk=record_pk,
        )

        src_record_get_or_create_key = (
            model_label,
            self._get_unique_preparing_model_record_key(
                record=src_record,
            ),
        )

        src_record_pk_key = (
            model_label,
            record_pk,
        )

        if src_record_get_or_create_key in self._preparing_result_model_records_map:  # noqa
            result_model_label, result_record_pk = (
                self._preparing_result_model_records_map[
                    src_record_get_or_create_key
                ]
            )

            result_record = self._result_storage.get_record(
                model_label=result_model_label,
                record_pk=result_record_pk,
            )
        else:
            result_model_record = deepcopy(src_record.model_record)

            for field_name, foreign_model_label in result_model_record.foreign_key_fields_map.items():  # noqa
                result_field_value = result_model_record.get_field_value(
                    name=field_name,
                )

                if result_field_value != 'None':
                    foreign_result_record_pk_key = (
                        foreign_model_label,
                        result_field_value.model_record.pk,
                    )

                    foreign_result_record = (
                        self._table_conformity[foreign_result_record_pk_key]
                    )

                    result_model_record.set_field_value(
                        name=field_name,
                        value=foreign_result_record,
                    )

                    foreign_result_record.has_dependence = True

            result_record = self._result_storage.add(
                model_record=result_model_record,
                tag=PreparingModelRecordTagEnum.CREATING,
                is_add_recursively_fks=True,
            )

        result_record.has_dependence = src_record.has_dependence

        self._table_conformity[src_record_pk_key] = result_record

    def _union(self):
        """
        Объединение записей хранилищ в одно общее хранилище
        """
        sorted_src_record_keys = self._sort_records(
            records=self._src_storage.records.values(),
        )

        for src_record_key in sorted_src_record_keys:
            self._transfer_src_record(
                src_record_key=src_record_key,
            )

    def run(self) -> PreparingModelRecordStorage:
        """
        Запуск действия на объединение записей хранилищ в единое хранилище
        """
        self._recalculate_src_storage_record_pks()
        self._prepare_result_storage()
        self._prepare_preparing_result_model_records_map()
        self._union()

        return self._result_storage



import codecs
import datetime
import os
import sys
from typing import (
    Optional,
)

from django.conf import (
    settings,
)

from factory_bo.enums import (
    DiffFormatEnum,
    FixtureDestinationEnum,
)
from factory_bo.exporters import (
    JSONModelExporter,
    PyFactoryExporter,
)
from factory_bo.helpers import (
    check_fixture_files,
    get_last_edited_fixtures,
)
from factory_bo.storages import (
    DefaultDBPreparingModelRecordStorageCreator,
    ExistingFactoryStorage,
    JSONPreparingModelStorageCreator,
    ModelFactoryCorrelationStorageCreator,
    PreparingModelRecordStorage,
)
from factory_bo.strings import (
    FACTORY_BO__FIXTURES_DIR_PATH_NOT_FOUND,
    WRONG_DIFF_FORMAT,
    WRONG_FIXTURE_FILE_EXTENSION_ERROR,
)


def get_preparing_model_record_storage_from_file(
    file_path: str,
    existing_factory_storage: ExistingFactoryStorage,
    fk_source_storage: Optional[PreparingModelRecordStorage] = None,
) -> PreparingModelRecordStorage:
    """
    Получение хранилища кешей фабрик на основе контента фикстур фабрик
    """
    with open(file_path, 'r') as f:
        factories_json = f.read()

    factory_cache_storage_creator = JSONPreparingModelStorageCreator(
        factories_json=factories_json,
        existing_factory_storage=existing_factory_storage,
    )

    return factory_cache_storage_creator.create(
        need_substitute_fk_ids=True,
        fk_source_storage=fk_source_storage,
    )


def create_models_json_fixture(
    fixture_destination: str = FixtureDestinationEnum.FILE,
    fixture_file_path: str = '',
):
    """
    Функция создания фикстуры записей моделей.

    Фикстура представлена в виде JSON-файла, содержащего в качестве ключа
    <app_name>.<model_class_name> и в качестве значения списка параметров для
    создания записей моделей.

    Вывод может осуществляться в стандартный поток вывода, либо в файл. За это
    поведение отвечает параметр fixture_destination. Принимает значения stdout
    или file.

    Если заранее известен абсолютный путь результирующего файла для выгрузки,
    то его можно указать с помощью параметра fixture_file_path. Аргумент
    должен заканчиваться на ".json". Если путь заранее не известен, то выгрузка
    будет произведена в файл FIXTURES_DIR_PATH.fixture_{%Y%m%d_%H%M%S}.json
    """
    if fixture_destination == FixtureDestinationEnum.FILE:
        if fixture_file_path:
            if fixture_file_path.endswith('.json'):
                file_path = fixture_file_path
            else:
                raise OSError(
                    WRONG_FIXTURE_FILE_EXTENSION_ERROR
                )
        else:
            now = datetime.datetime.now()
            file_path = os.path.join(
                settings.FACTORY_BO__FIXTURES_DIR_PATH,
                f'fixture_{now.strftime("%Y%m%d_%H%M%S")}.json'
            )

        dir_path = os.path.split(file_path)[0]

        os.makedirs(
            dir_path,
            exist_ok=True
        )
        f = codecs.open(file_path, 'w', encoding='utf-8')
    else:
        f = sys.stdout

    existing_factory_storage = ExistingFactoryStorage()

    preparing_model_storage_creator = (
        DefaultDBPreparingModelRecordStorageCreator(
            existing_factory_storage=existing_factory_storage,
        )
    )

    preparing_model_record_storage = (
        preparing_model_storage_creator.create()
    )

    json_model_exporter = JSONModelExporter(
        preparing_model_record_storage=preparing_model_record_storage,
        existing_factory_storage=existing_factory_storage,
    )

    content = json_model_exporter.to_string()

    f.write(content)
    f.close()


def create_diff_factories_fixture(
    begin_file_name: Optional[str] = None,
    end_file_name: Optional[str] = None,
    diff_format: str = DiffFormatEnum.PYTHON,
    result_file_path: Optional[str] = None,
):
    """
    Функция предназначена для получения предыстории тестов. Перед выполнением
    действий необходимо снять дамп эталонной базы в виде фикстуры с помощью
    команды make_factories_fixtures. После чего выполняются действия
    предыстории. Далее снимается второй дамп базы данных в виде фикстур.

    После чего выполняется данная функция, которая позволяет получить разницу
    между файлами с отсортированными фабриками и расставленными переменными
    вместо идентификаторов.

    Идентификаторы частично остаются из-за того, что объекты, на которые идет
    ссылка не создавались во время выполнения действий предыстории.

    Важно! В секции factory_bo конфигурационного файла нужно указать
    FIXTURES_DIR_PATH - абсолютный путь до директории, в которой
    хранятся фикстуры для тестирования. Начальная и конечная фиктуры должны
    быть созданы либо перенесены в данную директорию.

    Для работы функции могут быть указаны наименования фикстуры с фабриками
    полученные до и после выполнения действий предыстории с помощью параметров
    begin_factories_fixture_name и end_factories_fixture_name. Если эти
    параметры не будут указаны или будут заданы частично, то будет произведен
    поиск последних отредатированных фикстур фабрик и разница будет находиться
    между ними. Валидными являются фикстуры с контентом в формате JSON.

    Выгрузка может производиться в виде Python с полностью подготовленным кодом
    фикстуры фабрик для исполнения. Также есть возможность выгузки в виде
    SQL-запросов, полученных после исполнения Python-кода, полученного на
    предыщем шаге. Формат выгрузки можно указать с помощью параметра
    diff_format, который может принимать значение py или sql (выгрузка в
    формате SQL в данный момент не поддерживается).

    Если заранее известен абсолюный путь резульрирующего файла, то его можно
    указать с помощью параметра result_file_path. Данный параметр является не
    обязательным. Если он не указан, то выгрузка будет произведена в файл
    FIXTURES_DIR_PATH/diff_result/diff_fixtures__{%Y_%m_%d__%H_%M_%S}.{diff_format}.
    """
    if not settings.FACTORY_BO__FIXTURES_DIR_PATH:
        raise SystemExit(
            FACTORY_BO__FIXTURES_DIR_PATH_NOT_FOUND
        )

    if diff_format not in DiffFormatEnum.values:
        raise ValueError(
            WRONG_DIFF_FORMAT.format(str(diff_format))
        )

    if not begin_file_name or not end_file_name:
        begin_file_path, end_file_path = get_last_edited_fixtures()
    else:
        begin_file_path, end_file_path = check_fixture_files(
            begin_file_name=begin_file_name,
            end_file_name=end_file_name,
        )

    existing_factory_storage = ExistingFactoryStorage()

    begin_preparing_model_record_storage = (
        get_preparing_model_record_storage_from_file(
            file_path=begin_file_path,
            existing_factory_storage=existing_factory_storage,
        )
    )
    end_preparing_model_record_storage = (
        get_preparing_model_record_storage_from_file(
            file_path=end_file_path,
            fk_source_storage=begin_preparing_model_record_storage,
            existing_factory_storage=existing_factory_storage,
        )
    )

    difference_storage = end_preparing_model_record_storage.difference(
        begin_preparing_model_record_storage
    )

    correlation_storage_creator = ModelFactoryCorrelationStorageCreator(
        existing_factory_storage=existing_factory_storage,
    )

    correlation_storage = correlation_storage_creator.create()

    py_factory_exporter = PyFactoryExporter(
        preparing_model_record_storage=difference_storage,
        model_factory_correlation_storage=correlation_storage,
        existing_factory_storage=existing_factory_storage,
    )

    result_file_path = py_factory_exporter.to_file(
        file_path=result_file_path,
    )

    return result_file_path



from copy import (
    copy,
    deepcopy,
)
from typing import (
    Any,
    Dict,
    Optional,
    Set,
    Type,
)

from django.apps import (
    apps,
)
from django.contrib.contenttypes.fields import (
    GenericForeignKey,
)
from django.contrib.contenttypes.models import (
    ContentType,
)
from django.db.models import (
    Model,
)

from factory_bo.consts import (
    MODEL_CLASS_NAME_ID_DELIMITER,
)
from factory_bo.enums import (
    PreparingModelRecordTagEnum,
)
from factory_bo.signals import (
    prepare_record_foreign_key_fields_map_signal,
)


class ModelRecord:
    """
    Запись модели
    """

    def __init__(
        self,
        model_class: Type[Model],
        foreign_key_fields_map: Dict[str, str],
        fields: Dict[str, Any],
        content_type_cache: Dict[int, ContentType],
    ):
        self._model_class: Type[Model] = model_class
        self._fields = fields

        # Если есть поле id, то проставляется значение из него и удаляется из
        # списка полей, если нет, то присваивается значение первичного ключа
        # модели. Первичным ключом может выступать внешний ключ и тогда поле
        # не нужно убирать из списка полей.
        pk = self._model_class._meta.pk
        self._pk = (
            self._fields[pk.attname] if (
                pk.is_relation or
                # подразумевается, что может прилететь UUID в виде строки
                not self._fields[pk.attname].isdigit()
            ) else
            self._fields.pop(pk.attname, None)
        )

        # На случай, если какое-то поле кроме id является первичным ключом
        if 'id' in self._fields.keys():
            self._fields.pop('id')

        # Копируем foreing_key_field_map для дальнейшего расширения
        # GeneficForegnKey и прочих
        self._foreign_key_fields_map: Dict[str, str] = deepcopy(
            foreign_key_fields_map
        )

        self._prepare_foreign_key_fields_map(content_type_cache)

        self._fk_field_names = self._prepare_fk_field_names()

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'@model_label="{self.model_label}" '
            f'@pk={self._pk}>'
        )

    def __str__(self):
        return self.__repr__()

    @property
    def pk(self):
        return self._pk

    @pk.setter
    def pk(
        self,
        value,
    ):
        self._pk = str(value)

    @property
    def all_fields(self) -> Dict[str, str]:
        """
        Возвращает все поля фабрики в виде словаря
        """
        fields = copy(self.fields)
        fields[self._model_class._meta.pk.attname] = (
            self._pk
        )

        return fields

    @property
    def fields(self) -> Dict[str, str]:
        return self._fields

    @property
    def model_label(self) -> str:
        return self._model_class._meta.label

    @property
    def model_class(self) -> Type[Model]:
        return self._model_class

    @property
    def key(self) -> tuple:
        return (
            self.model_label,
            self.pk,
        )

    @property
    def foreign_key_fields_map(self) -> Dict[str, str]:
        return self._foreign_key_fields_map

    @property
    def fk_field_names(self):
        """
        Возвращает исчерпывающий список имен полей внешних ключей
        """
        return self._fk_field_names

    @property
    def model_label_with_pk(self):
        """
        Возвращает строку с названием модели и id записи разделённые специальным разделителем.
        """
        return f'{self.model_label}{MODEL_CLASS_NAME_ID_DELIMITER}{self.pk}'

    def get_field_value(
        self,
        name: str,
    ):
        """
        Возвращает значение поля по имени
        """
        return self._fields[name]

    def set_field_value(
        self,
        name: str,
        value: str,
    ):
        """
        Установка значения поля у записи
        """
        self._fields[name] = value

    def _prepare_generic_foreign_key(
        self,
        generic_foreign_key: GenericForeignKey,
        content_type_cache: Dict[int, ContentType],
    ):
        """
        Подготовка GenericForeignKey для дальнейшего использования в
        Python-фикстурах
        """
        content_type_field_name = (
            generic_foreign_key.ct_field if
            generic_foreign_key.ct_field.endswith('_id') else
            f'{generic_foreign_key.ct_field}_id'
        )
        content_type_id = self.fields.get(content_type_field_name)
        object_id = self.fields.get(generic_foreign_key.fk_field)

        if content_type_id != 'None' and object_id != 'None':
            content_type = content_type_cache.get(int(content_type_id))
            model = apps.get_model(
                content_type.app_label,
                content_type.model
            )
            self._foreign_key_fields_map[generic_foreign_key.fk_field] = (
                model._meta.label
            )

    def _prepare_generic_foreign_keys(
        self,
        content_type_cache: Dict[int, ContentType],
    ):
        """
        Подготовка GenericForeignKey для дальнейшего использования в
        Python-фикстурах

        Подготовка подразумевает изменение типа поля object_id к ForeignKey для
        корректной подстановки фабрики. Это необходимо, т.к. при изменении
        идентификаторов в эталонной БД, например, при ее перегенерации
        """
        generic_foreign_keys = set(
            filter(
                lambda field: isinstance(field, GenericForeignKey),
                self.model_class._meta.private_fields
            )
        )

        for generic_foreign_key in generic_foreign_keys:
            self._prepare_generic_foreign_key(
                generic_foreign_key=generic_foreign_key,
                content_type_cache=content_type_cache,
            )

    def _prepare_foreign_key_fields_map(
        self,
        content_type_cache: Dict[int, ContentType],
    ):
        """
        Подготовка карты соответствия внешних ключей и моделей

        Необходимо определять карту для каждой записи, потому что они могут
        содержать GenericForeignKey или другие кастомные структуры, которые
        явно не являются внешними ключами, но подразумевают такое поведение
        """
        self._prepare_generic_foreign_keys(content_type_cache)

        prepare_record_foreign_key_fields_map_signal.send(
            sender=self,
        )

    def _prepare_fk_field_names(self):
        """
        Возвращает исчерпывающий список имен полей внешних ключей
        """
        return list(self._foreign_key_fields_map.keys())


class PreparingModelRecord:
    """
    Обрабатываемая запись модели
    """

    def __init__(
        self,
        model_record: ModelRecord,
        tag: int,
        changed_fields: Optional[Dict[str, Any]] = None,
        has_dependence: bool = False,
    ):
        self._model_record: ModelRecord = model_record
        self._tags: Set[int] = set()
        self._changed_fields: Optional[Dict[str, Any]] = None
        self._has_dependence = has_dependence

        self.set_tag(
            tag=tag,
            changed_fields=changed_fields,
        )

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} @model_record="{self._model_record}" '
            f'@tags="{", ".join(map(str, self._tags))}">'
        )

    def __str__(self):
        return self.__repr__()

    @property
    def model_record(self) -> ModelRecord:
        return self._model_record

    @model_record.setter
    def model_record(self, model_record):
        self._model_record = model_record

    @property
    def tags(self) -> Set[int]:
        return self._tags

    @property
    def has_getting_tag(self):
        return PreparingModelRecordTagEnum.GETTING in self.tags

    @property
    def has_only_getting_tag(self):
        return (
            self.has_getting_tag and
            not self.has_creating_tag and
            not self.has_updating_tag and
            not self.has_deleting_tag
        )

    @property
    def has_creating_tag(self):
        return PreparingModelRecordTagEnum.CREATING in self.tags

    @property
    def has_no_creating_tag(self):
        return not self.has_creating_tag and (
            self.has_getting_tag or
            self.has_updating_tag or
            self.has_deleting_tag
        )

    @property
    def has_updating_tag(self):
        return PreparingModelRecordTagEnum.UPDATING in self.tags

    @property
    def has_deleting_tag(self):
        return PreparingModelRecordTagEnum.DELETING in self.tags

    @property
    def changed_fields(self) -> Optional[Dict[str, Any]]:
        return self._changed_fields

    @property
    def has_dependence(self):
        return self._has_dependence

    @has_dependence.setter
    def has_dependence(
        self,
        has_dependence: bool = False,
    ):
        self._has_dependence = has_dependence

    def set_tag(
        self,
        tag: int,
        changed_fields: Optional[Dict[str, Any]] = None,
    ):
        """
        Установка метки записи обратываемой записи. Метки влияют на действия с
        записями при выгрузке

        Если тег будет равен PreparingModelRecordTagEnum.UPDATING, то
        дополнительно устанавливаются измененные поля changed_fields

        Если у записи уже установлен тег из несовместимых
        PreparingModelRecordTagEnum.INCOMPATIBLES и совершается попытка
        добавить еще один, то об этом нужно оповестить пользователя
        """
        tag_is_incompatible = PreparingModelRecordTagEnum.is_incompatible(
            tag=tag,
        )

        if tag not in self._tags:
            if (
                tag_is_incompatible and
                PreparingModelRecordTagEnum.have_incompatible(self._tags)
            ):
                raise ValueError(
                    f'Can not set "{tag}" tag, because {self} already have '
                    f'incompatible tag!'
                )

            if (
                PreparingModelRecordTagEnum.is_updating(tag) and
                not changed_fields
            ):
                raise ValueError(
                    f'If you set UPDATING tag, you should send a changed '
                    f'fields dictionary.'
                )

            self._tags.add(tag)

            if PreparingModelRecordTagEnum.is_updating(tag):
                self._changed_fields = changed_fields

            if (
                PreparingModelRecordTagEnum.is_deleting(tag) or
                PreparingModelRecordTagEnum.is_updating(tag)
            ):
                self._tags.add(PreparingModelRecordTagEnum.GETTING)



from django.core.management import (
    BaseCommand,
)

from factory_bo.api import (
    create_models_json_fixture,
)
from factory_bo.enums import (
    FixtureDestinationEnum,
)


class Command(BaseCommand):
    """
    Команда создания фикстуры записей моделей.

    Фикстура представлена в виде JSON-файла, содержащего в качестве ключа
    <app_name>.<model_class_name> и в качестве значения списка параметров для
    создания записей моделей.

    Вывод может осуществляться в стандартный поток вывода, либо в файл. За это
    поведение отвечает параметр --fixture_destination. Принимает значения stdout
    или file.

    Если заранее известен абсолютный путь результирующего файла для выгрузки,
    то его можно указать с помощью параметра --fixture_file_path. Аргумент
    должен заканчиваться на ".json". Если путь заранее не известен, то выгрузка
    будет произведена в файл FIXTURES_DIR_PATH.fixture_{%Y%m%d_%H%M%S}.json
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '--fixture_destination',
            action='store',
            dest='fixture_destination',
            default=FixtureDestinationEnum.FILE,
            help=(
                'Output method. You can choose stdout or file. Default is '
                'stdout.'
            )
        )

        parser.add_argument(
            '--fixture_file_path',
            action='store',
            dest='fixture_file_path',
            help=(
                'Absolute path of fixture json file.'
            )
        )

    def handle(self, *args, **options):
        fixture_destination = options['fixture_destination']
        fixture_file_path = options['fixture_file_path'] or ''

        create_models_json_fixture(
            fixture_destination=fixture_destination,
            fixture_file_path=fixture_file_path,
        )



from typing import (
    List,
    Set,
    Tuple,
)

from django.core.management import (
    BaseCommand,
)

from factory_bo.enums import (
    PreparingModelRecordTagEnum,
)
from factory_bo.exporters import (
    HierarchyForRecordsPyFactoryExporter,
)
from factory_bo.records import (
    PreparingModelRecord,
)
from factory_bo.storages import (
    DefaultDBPreparingModelRecordStorageCreator,
    ExistingFactoryStorage,
    ModelFactoryCorrelationStorageCreator,
    PreparingModelRecordStorage,
)


class Command(BaseCommand):
    """
    Формирует иерархию фабрик для переданных записей моделей и выводит в консоль.
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '--model-label-with-id',
            action='store',
            dest='model_label_with_id',
            required=True,
            help='List of django models with id. For example: repos.DocumentTypes__1234.',
        )
        parser.add_argument(
            '--with-imports',
            action='store_true',
            dest='with_imports',
            default=False,
            help='Include python imports for factories in output.'
        )

    def _parse_model_label_with_id(
        self,
        options: dict,
    ) -> List[Tuple[str, str]]:
        """Возвращает список кортежей, содержащих пары значений - Наименование модели, Идентификатор."""

        return [tuple(record.split('__')) for record in options['model_label_with_id'].split(',')]

    def _get_preparing_model_records(
        self,
        storage: PreparingModelRecordStorage,
        model_label_with_id: List[Tuple[str, str]],
    ) -> List[PreparingModelRecord]:
        """Получает список обработанных записей моделей из хранилища.

        Args:
            storage: Хранилище обрабатываемых записей моделей.
            model_label_with_id: Список кортежей, содержащих пары значений - Наименование модели, Идентификатор.

        Returns:
            Список обработанных записей моделей.

        """
        result = []

        for model_label, record_id in model_label_with_id:
            try:
                preparing_model_record = storage.get_record(model_label, record_id)
            except KeyError as e:
                print(f'Для модели "{model_label}" не найдено записи с id "{record_id}"')

                raise e

            result.append(preparing_model_record)

        return result

    def handle(self, *args, **options):
        model_label_with_id = self._parse_model_label_with_id(options)
        with_imports = options['with_imports']

        existing_factory_storage = ExistingFactoryStorage(
            raise_on_nonexistent=False,
        )

        model_record_storage_creator = DefaultDBPreparingModelRecordStorageCreator(
            existing_factory_storage=existing_factory_storage,
        )
        preparing_model_record_storage = model_record_storage_creator.create(
            need_substitute_fk_ids=True,
            preparing_model_record_tag=PreparingModelRecordTagEnum.GETTING,
        )

        initial_preparing_model_records = self._get_preparing_model_records(
            preparing_model_record_storage,
            model_label_with_id,
        )

        preparing_model_records: Set[PreparingModelRecord] = set()
        for record in initial_preparing_model_records:
            dependencies = preparing_model_record_storage.get_dependencies_for_record_by_fk_recursively(record)
            preparing_model_records.update(dependencies)

        preparing_model_records.update(initial_preparing_model_records)

        model_factory_correlation_storage = ModelFactoryCorrelationStorageCreator(
            existing_factory_storage=existing_factory_storage,
        ).create()

        exporter = HierarchyForRecordsPyFactoryExporter(
            model_factory_correlation_storage=model_factory_correlation_storage,
            preparing_model_record_storage=preparing_model_record_storage,
            existing_factory_storage=existing_factory_storage,
        )
        result_string = exporter.to_string(preparing_model_records, with_imports)

        print(f'Для записей {model_label_with_id} сформирована иерархия фабрик:\n\n{result_string}')



from collections import (
    defaultdict,
)

from django.apps import (
    apps,
)
from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
)


class Command(BaseCommand):
    """
    Проверяет целостность данных, в части использования self fk полей
    """

    def handle(self, *args, **options):
        SELF_FK_IDS = (
            'external_id',
            'direct_id',
            'root_id',
            'parent_id',
        )

        models = apps.get_models(include_auto_created=True)
        model_self_fks = defaultdict(dict)

        for model in models:
            if 'web-bb-salary' in model._meta.app_config.path:
                model.objects = model._base_manager

                fields_list = (
                    field.attname
                    for field in model._meta.fields
                )

                self_fks = set(SELF_FK_IDS).intersection(fields_list)

                if self_fks:
                    for self_fk in self_fks:
                        model_self_fks[model._meta.label][self_fk] = set()

                        self_fk_ids = set(
                            model.objects.exclude(
                                **{
                                    f'{self_fk}__isnull': True,
                                }
                            ).values_list(self_fk, flat=True)
                        )

                        if self_fk_ids:
                            for self_fk_id in self_fk_ids:
                                try:
                                    model.objects.get(
                                        **{
                                            'pk': self_fk_id,
                                        }
                                    )
                                except model.DoesNotExist:
                                    model_self_fks[model._meta.label][self_fk].add(
                                        self_fk_id
                                    )

        file_path = (
            f'{settings.FACTORY_BO__FIXTURES_DIR_PATH}/'
            f'salary_db_self_fks_consistency.csv'
        )
        with open(file_path, 'w') as f:
            for model_label, self_fks_dict in model_self_fks.items():
                for self_fk, self_fk_ids in self_fks_dict.items():
                    if self_fk_ids:
                        f.write(f'{model_label},{self_fk}\n')
                        for self_fk_id in self_fk_ids:
                            f.write(f',{self_fk_id}\n')



TAB_STR = '    '

MODEL_CLASS_NAME_ID_DELIMITER = FACTORY_CLASS_NAME_DELIMETER = '___'

IMPORT_TEMPLATE = 'from {import_path} import {class_name}'
IMPORT_WITH_ALIAS_TEMPLATE = 'from {import_path} import {class_name} as {class_name_alias}'  # noqa

EXCLUDED_FIELD_CLASSES = (
    'FileField',
)

GENERATED_PREFIX = 'Generated'
ACTUAL_PREFIX = 'Actual'
GENERATED_ACTUAL_PREFIX = 'GeneratedActual'

CONTENT_TYPE_MODEL = 'contenttypes.ContentType'


# Сдвиг значений первичных ключей записей сорцового хранилища при объединении
# двух хранилищ
PK_OFFSET = 100_000_000



from django.core.management import (
    BaseCommand,
)

from factory_bo.api import (
    create_diff_factories_fixture,
)
from factory_bo.enums import (
    DiffFormatEnum,
)


class Command(BaseCommand):
    """
    Команда предназначена для получения предыстории тестов. Перед выполнением
    действий необходимо снять дамп эталонной базы в виде фикстуры с помощью
    команды make_factories_fixtures. После чего выполняются действия
    предыстории. Далее снимается второй дамп базы данных в виде фикстур.

    После чего выполняется данная команда, которая позволяет получить разницу
    между файлами с отсортированными фабриками и расставленными переменными
    вместо идентификаторов.

    Идентификаторы частично остаются из-за того, что объекты, на которые идет
    ссылка не создавались во время выполнения действий предыстории.

    Важно! В секции factory_bo конфигурационного файла нужно указать
    FIXTURES_DIR_PATH - абсолютный путь до директории, в которой
    хранятся фикстуры для тестирования. Начальная и конечная фиктуры должны
    быть созданы либо перенесены в данную директорию.

    Для работы команды могут быть указаны наименования фикстуры с фабриками
    полученные до и после выполнения действий предыстории с помощью параметров
    --begin_factories_fixture_name и --end_factories_fixture_name. Если эти
    параметры не будут указаны или будут заданы частично, то будет произведен
    поиск последних отредатированных фикстур фабрик и разница будет находиться
    между ними. Валидными являются фикстуры с контентом в формате JSON.

    Выгрузка может производиться в виде Python с полностью подготовленным кодом
    фикстуры фабрик для исполнения. Также есть возможность выгузки в виде
    SQL-запросов, полученных после исполнения Python-кода, полученного на
    предыщем шаге. Формат выгрузки можно указать с помощью параметра
    --diff_format, который может принимать значение py или sql.

    Если заранее известен абсолюный путь резульрирующего файла, то его можно
    указать с помощью параметра --result_file_path. Данный параметр является не
    обязательным. Если он не указан, то выгрузка будет произведена в файл
    FIXTURES_DIR_PATH/diff_result/diff_fixtures__{%Y_%m_%d__%H_%M_%S}.{diff_format}.

    Пример запуска:
    python manage.py make_diff_fixtures --begin_factories_fixture_name begin.py --end_factories_fixture_name end.py  --diff_format=py --result_file_path=/some_path/some_file.py
    """
    def add_arguments(self, parser):
        parser.add_argument(
            '--begin_factories_fixture_name',
            action='store',
            dest='begin_factories_fixture_name',
            help='Factories fixture name before actions',
        )

        parser.add_argument(
            '--end_factories_fixture_name',
            action='store',
            dest='end_factories_fixture_name',
            help='Factories fixture name after actions',
        )

        parser.add_argument(
            '--diff_format',
            action='store',
            dest='diff_format',
            default=DiffFormatEnum.PYTHON,
            help=(
                'Diff export format. You can choose py or sql format. '
                'Default is py.'
            ),
        )

        parser.add_argument(
            '--result_file_path',
            action='store',
            dest='result_file_path',
            default=None,
            help=(
                'File path for exporting results.'
            ),
        )

    def handle(self, *args, **options):
        begin_file_name = options.get('begin_factories_fixture_name')
        end_file_name = options.get('end_factories_fixture_name')
        diff_format = options.get('diff_format')
        result_file_path = options.get('result_file_path')

        result_file_path = create_diff_factories_fixture(
            begin_file_name=begin_file_name,
            end_file_name=end_file_name,
            diff_format=diff_format,
            result_file_path=result_file_path,
        )

        self.stdout.write(
            f'Difference fixtures result file path - {result_file_path}'
        )



import re
from collections import (
    defaultdict,
)
from typing import (
    List,
    Optional,
)

from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
)
from django.db.models import (
    Model,
)
from isort.api import (
    sort_code_string,
)

from factory_bo.consts import (
    ACTUAL_PREFIX,
    GENERATED_PREFIX,
    TAB_STR,
)
from factory_bo.factories import (
    ExistingFactory,
)
from factory_bo.storages import (
    ExistingFactoryStorage,
)
from factory_bo.strings import (
    FACTORY_CLASS_WITHOUT_COMMENT_STR,
    IMPORT_DEFAULT_MANAGER_FACTORY_STR,
    IMPORT_FACTORY_USE_TYPE_ENUM_STR,
)


class Command(BaseCommand):
    """
    Команда для негерации фабрик для моделей, у которых отсутствует дефолтная
    фабрика

    Для моделей, не относящихся к проектам, фабрики будут генерировать в
    директории

    FACTORY_BO__FIXTURES_DIR_PATH/features/factories/{app_label}.py
    """

    def _generate_factory_class_name(
        self,
        model: Model,
    ) -> str:
        """
        Генерация имени фабрики для модели
        """
        model_class_name = model._meta.label.split('.')[1]

        model_class_name_parts = list(
            map(
                lambda part: (
                    part.title() if
                    part.islower() else
                    part
                ),
                model_class_name.split('_')
            )
        )

        return f'GeneratedActual{"".join(model_class_name_parts)}Factory'

    def _write_imports(
        self,
        content: str,
    ) -> str:
        """
        Запись в файл импорта базового класа фабрик DefaultManagerFactory
        """
        imports = []

        if 'DefaultManagerFactory' in content:
            imports.append(IMPORT_DEFAULT_MANAGER_FACTORY_STR)

        return '\n'.join([content, *imports])

    def _sort_imports(
        self,
        content: str,
    ) -> str:
        """
        Сортировка импортов
        """
        return sort_code_string(
            code=content,
            config=settings.ISORT_CONFIG,
        )

    def _get_model_get_or_create_fields(
        self,
        model: Model,
        existing_factory: Optional[ExistingFactory] = None
    ) -> str:
        """
        Получение списка полей django_get_or_create в виде строки с
        перечислением через запятую
        """
        model_field_names = [
            field.name
            for field in model._meta.fields
        ]
        django_get_or_create_fields = set()

        if 'name' in model_field_names:
            django_get_or_create_fields.add('name')

        django_get_or_create_fields.update(
            set(model_field_names).intersection(
                settings.FACTORY_BO__PSEUDO_SELF_FK_IDS
            )
        )

        django_get_or_create_fields.update(
            set(model_field_names).intersection(
                settings.FACTORY_BO__PSEUDO_FK_IDS
            )
        )

        if existing_factory:
            field_names = (
                existing_factory.factory_class.get_django_get_or_create()
            )
            django_get_or_create_fields.update(
                [
                    field_name
                    for field_name in field_names if
                    '_id' not in field_name
                ]
            )

        django_get_or_create_fields_str = f'{TAB_STR*2}'
        if django_get_or_create_fields:
            django_get_or_create_fields_str = ',\n'.join(
                sorted(
                    map(
                        lambda field_str: f'{TAB_STR*3}\'{field_str}\'',  # noqa
                        django_get_or_create_fields
                    )
                )
            )
            django_get_or_create_fields_str = (
                f'{django_get_or_create_fields_str},'
            )

        return django_get_or_create_fields_str

    def _generate_actual_factory_for_model(
        self,
        content: str,
        existing_factories: List[ExistingFactory],
    ) -> str:
        generated_entities = []

        for existing_factory in existing_factories:
            model = existing_factory.factory_class._meta.model

            django_get_or_create_str = (
                self._get_model_get_or_create_fields(
                    model=model,
                    existing_factory=existing_factory,
                )
            )

            factory_class_name = self._generate_factory_class_name(
                model=model,
            )

            base_factory_class = 'DefaultManagerFactory'

            generated_entities.append(
                FACTORY_CLASS_WITHOUT_COMMENT_STR.format(
                    base_factory_class=base_factory_class,
                    factory_class_name=factory_class_name,
                    model_label=model._meta.label,
                    django_get_or_create_str=django_get_or_create_str,
                )
            )

        return ''.join([content, *generated_entities])

    def _mark_default_factories_as_custom(
        self,
        module_path: str,
        factory_class_names,
    ) -> str:
        """
        В файлах фабрик производится пометка дефолтной фабрики, как кастомной
        для дальнейшей генерации актуальной, на ее основе
        """
        code_lines = []
        with open(module_path) as original_file:
            need_add_blank_line = False
            for line in original_file:
                if any(map(lambda name: f'class {name}(' in line, factory_class_names)):
                    line = re.sub(
                        r'(class [\w\d_]+Factory\([\w\d_]+\):)',
                        r'\g<1>\n    _factory_use_type = FactoryUseTypeEnum.CUSTOM',  # noqa
                        line
                    )

                    need_add_blank_line = True
                else:
                    if need_add_blank_line and line and line != '\n':
                        line = f'\n{line}'

                    need_add_blank_line = False

                code_lines.append(line)

        code_lines.append(IMPORT_FACTORY_USE_TYPE_ENUM_STR)

        code = ''.join(code_lines)

        return code

    def handle(self, *args, **options):
        existing_factory_storage = ExistingFactoryStorage()

        model_default_factory_map = (
            existing_factory_storage.get_model_default_factory_map()
        )

        factory_classes = defaultdict(set)

        for existing_factory in model_default_factory_map.values():
            if (
                existing_factory.factory_class_name.startswith(GENERATED_PREFIX) or  # noqa
                existing_factory.factory_class_name.startswith(ACTUAL_PREFIX)
            ):
                continue

            factory_classes[existing_factory.module_path].add(existing_factory)

        for module_path, existing_factories in factory_classes.items():
            if 'web-bb-salary' not in module_path:
                continue

            factory_class_names = list(
                existing_factory.factory_class_name
                for existing_factory in existing_factories
            )
            code = self._mark_default_factories_as_custom(
                module_path=module_path,
                factory_class_names=factory_class_names,
            )

            code = self._generate_actual_factory_for_model(
                content=code,
                existing_factories=existing_factories,
            )

            code = self._write_imports(
                content=code,
            )

            code = self._sort_imports(
                content=code,
            )

            with open(module_path, 'w') as original_file:
                original_file.write(code)



import fileinput
import inspect
import re
from copy import (
    deepcopy,
)
from importlib import (
    import_module,
)

from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
)
from django.db import (
    models,
)
from django.db.models import (
    Count,
)

from factory_bo.consts import (
    GENERATED_PREFIX,
    TAB_STR,
)
from factory_bo.storages import (
    ExistingFactoryStorage,
)


class Command(BaseCommand):
    """
    Команда проверки набора полей django_get_or_create у фабрик
    на уникальность выборки, и замены набора полей на те,
    которые дают уникальную выборку у соответствующей модели.
    """

    excluded_fields = (
        'id',
        'version',
    )
    min_records_count = 2

    def add_arguments(self, parser):
        parser.add_argument(
            '--check_generated_factories',
            action='store_true',
            dest='check_generated_factories',
            default=False,
            help='Check only factories with Generated-prefix',
        )
        parser.add_argument(
            '--update_get_or_create_fields',
            action='store_true',
            dest='update_get_or_create_fields',
            default=False,
            help='Update factories django_get_or_create fields',
        )
        parser.add_argument(
            '--remove_generated_prefix',
            action='store_true',
            dest='remove_generated_prefix',
            default=False,
            help='Remove factory name Generated-prefix',
        )
        parser.add_argument(
            '--another_db_settings',
            dest='another_db_settings',
            default=None,
            help=(
                'Use database-settings to route another DB. '
                'Format: NAME|USER|PASSWORD|HOST|PORT'
            ),
        )
        parser.add_argument(
            '--factories',
            dest='factories',
            default=None,
            help=(
                'Collect get_or_create fields '
                'for one or many specified factories. '
                'Format: SomeFactoryClassName|AnotherFactoryClassName'
            ),
        )
        parser.add_argument(
            '--apps_list',
            dest='apps_list',
            default=None,
            help=(
                'Collect get_or_create fields for one or many specified apps. '
                'Format: some_app_name,another_app_name'
            ),
        )

    def log(self, message):
        print(message)

    def _extend_db_settings(self, options):
        """
        Расщирение настроек БД.
        Формат строки настроек NAME|USER|PASSWORD|HOST|PORT.
        """
        another_db_settings = options.get('another_db_settings')

        if another_db_settings:
            db_settings = another_db_settings.split('|')

            default_db_settings = deepcopy(settings.DATABASES['default'])
            default_db_settings.update({
                'NAME': db_settings[0],
                'USER': db_settings[1],
                'PASSWORD': db_settings[2],
                'HOST': db_settings[3],
                'PORT': db_settings[4],
            })
            settings.DATABASES['another_db'] = default_db_settings

    def _prepare_model_factory_map(self, raise_on_nonexistent=True) -> dict:
        """Подготавливает хранилище фабрик, а так же карты моделей.

        Args:
            raise_on_nonexistent: Вызвать исключение при отсутствии дефолтной фабрики модели.

        Returns:
            Карты соответствия моделей и их дефолтных фабрик.
        """
        existing_factory_storage = ExistingFactoryStorage(raise_on_nonexistent=raise_on_nonexistent)
        self.models_map = existing_factory_storage._model_storage._models

        return existing_factory_storage.get_model_default_factory_map()

    def _get_model_allowed_fields(self, model_label):
        """
        Получение допустимых полей модели.
        """
        return {
            field
            for field in self.models_map[model_label]['allowed_fields']
            if field not in self.excluded_fields
        }

    @classmethod
    def _check_fk_key(cls, field):
        """
        Проверка что поле - внешний ключ
        """
        return (
            isinstance(field, models.IntegerField) and
            field.attname.endswith('_id')
        ) or isinstance(field, (models.ForeignKey, models.OneToOneField))

    @classmethod
    def _get_field_priority(cls, field):
        """
        Вычисление приоритета поля в зависимости от типа или названия
        """
        if cls._check_fk_key(field):
            priority = 1
        elif isinstance(field, (models.DateField, models.DateTimeField)):
            priority = 2
        elif field.attname in {'code', 'name', 'number', 'state'}:
            priority = 3
        else:
            priority = 4

        return priority

    def _prepare_get_or_create_fields(self, factory, model_allowed_fields):
        """
        Если у фабрики есть поля get_or_create - используем их,
        а в качестве остальных полей берём разницу между
        полями модели (model_allowed_fields) и полями get_or_create фабрики.

        Если у фабрики нет полей get_or_create. То начальный набор полей
        формируем по полям внешник ключей (заканчиваются на _id),
        а в качестве остальных полей берём остальные поля модели.
        """
        get_or_create_fields = set(factory.get_or_create_fields)

        model_id_fields = {
            f.attname for f in factory.factory_model._meta.fields
            if self._check_fk_key(f) and f.attname in model_allowed_fields
        }

        if get_or_create_fields:
            another_fields = model_allowed_fields.difference(get_or_create_fields)
        else:
            get_or_create_fields = model_id_fields
            another_fields = {
                f for f in model_allowed_fields
                if f not in model_id_fields
            }

        another_fields = [
            f.attname
            for f in sorted(
                factory.factory_model._meta.fields,
                key=self._get_field_priority,
            )
            if f.attname in another_fields
        ]

        if not get_or_create_fields and another_fields:
            get_or_create_fields = {another_fields.pop(0)}

        return get_or_create_fields, another_fields

    def _enough_model_records_count(self, model, db_key):
        """
        Проверка достаточного количества записей модели
        """
        enough_records = True

        try:
            if model.objects.using(db_key).count() < self.min_records_count:
                self.log(f'\tNot enough records ({db_key}).')
                enough_records = False
        except Exception as e:
            self.log(str(e))
            enough_records = False

        return enough_records

    def _get_duplicates(self, db_key, get_or_create_fields, model):
        """
        Для конкретной БД, группировка записей модели по указанному набору полей
        с подсчётом количества и последующей фильтрацией результата
        по количеству, для выявлениня дублирующихся записей.
        """
        duplicate_records = model.objects.using(db_key).values(
            *get_or_create_fields
        ).annotate(
            duplicates_count=Count(model._meta.pk.column),
        ).filter(
            duplicates_count__gt=1,
        )

        return duplicate_records

    def _get_class_name(self, line):
        """
        Предпринимает попытку вернуть наименование класса из пришедшей строки
        из файла
        """
        class_name = ''

        class_regex = re.match(r'class (\w*)\(\w*\):', line.strip())
        if class_regex:
            class_name = class_regex.group(1)

        return class_name

    def _update_factory_fields(self, factory, get_or_create_fields, has_duplicates):
        """
        Заменить или добавить набор get_or_create-полей в класс фабрики
        """
        module_object = import_module(factory.import_path)
        factory_class = getattr(module_object, factory.factory_class_name)
        class_lines = inspect.getsourcelines(factory_class)[0]
        class_str = ''.join(class_lines)
        class_matched = False

        for line in fileinput.input([factory.module_path], inplace=True):
            if factory.factory_class_name == self._get_class_name(line):
                class_matched = True
                class_lines.pop(0)

                fields = "".join([
                    f"{TAB_STR*3}'{f}',\n"
                    for f in sorted(get_or_create_fields)
                ])
                attribute = 'django_get_or_create'
                fields = f'{attribute} = (\n{fields}{TAB_STR*2})'

                if has_duplicates:
                    fields = (
                        f'# Подобранный набор полей может не обеспечивать'
                        f' уникальность выборки\n{TAB_STR*2}{fields}'
                    )

                if attribute in class_str:
                    class_str = re.sub(
                        fr'{attribute}.+\(([^)]+)\)',
                        fields,
                        class_str,
                    )

                    if f'# {attribute}' in class_str:
                        class_str = class_str.replace(
                            f'# {attribute}',
                            f'{attribute}',
                        )
                else:
                    class_str = re.sub(
                        '(model = .+\n)',
                        f'\\g<1>{TAB_STR*2}{fields}\n',
                        class_str,
                    )

                if self.remove_generated_prefix:
                    class_str = class_str.replace(
                        f'class {GENERATED_PREFIX}',
                        'class '
                    )

                print(class_str, end='')
            elif class_matched and line in class_lines:
                class_lines.pop(0)
            else:
                print(line, end='')

    def handle(self, *args, **options):
        check_generated_factories = options.get('check_generated_factories')
        update_get_or_create_fields = options.get('update_get_or_create_fields')
        self.remove_generated_prefix = options.get('remove_generated_prefix')

        if options.get('factories'):
            specified_factories = options.get('factories').split('|')
        else:
            specified_factories = None

        self._extend_db_settings(options)

        if options.get('apps_list'):
            apps_list = options.get('apps_list').split(',')
            raise_on_nonexistent = False
        else:
            raise_on_nonexistent = True

        model_default_factory_map = self._prepare_model_factory_map(raise_on_nonexistent)

        for model_label, factory in model_default_factory_map.items():
            if model_label.split('.')[0] not in apps_list:
                continue

            factory_class_name = factory.factory_class.__name__

            if (
                specified_factories and
                factory_class_name not in specified_factories
            ) or (
                check_generated_factories and
                not factory_class_name.startswith(GENERATED_PREFIX)
            ):
                # пропуск уже актуализированных фабрик
                continue

            model = factory.factory_class.get_model()
            model.objects = model._base_manager

            self.log(f'\n{factory.factory_class_name} - {model_label}')

            model_allowed_fields = self._get_model_allowed_fields(
                model_label
            )
            
            if not model_allowed_fields:
                self.log('\tModel has no "allowed_fields".')
                continue

            factory_get_or_create_fields, another_fields = self._prepare_get_or_create_fields(
                factory,
                model_allowed_fields,
            )

            if model._meta.unique_together:
                get_or_create_fields = {
                    f.attname
                    for f in model._meta.fields
                    if f.name in model._meta.unique_together[0]
                }

                has_duplicates = False
            else:
                get_or_create_fields = factory_get_or_create_fields

                db_duplicates = {}

                for db_key in settings.DATABASES:

                    if not self._enough_model_records_count(model, db_key):
                        continue

                    duplicates_exists = True
                    another_fields_exists = True

                    while duplicates_exists and another_fields_exists:
                        duplicate_records = self._get_duplicates(
                            db_key,
                            get_or_create_fields,
                            model,
                        )
                        duplicates_exists = duplicate_records.exists()

                        if duplicates_exists and another_fields:
                            get_or_create_fields.add(another_fields.pop(0))
                        elif not another_fields:
                            another_fields_exists = False

                    if duplicates_exists:
                        self.log(
                            f'\tDuplicate records found! ({db_key}) '
                            f'django_get_or_create - '
                            f'{", ".join(get_or_create_fields)}'
                        )
                        self.log(f'\t\t{duplicate_records.query}')
                    else:
                        self.log(
                            f'\tNo duplicates found! ({db_key}) '
                            f'django_get_or_create - '
                            f'{", ".join(get_or_create_fields)}'
                        )

                    db_duplicates[db_key] = duplicates_exists

                has_duplicates = any(db_duplicates.values())

            if (
                get_or_create_fields and
                    update_get_or_create_fields and
                    factory.get_or_create_fields != get_or_create_fields
            ):
                self._update_factory_fields(
                    factory,
                    get_or_create_fields,
                    has_duplicates,
                )
            elif not get_or_create_fields:
                self.log('\tDjango_get_or_create-fields was not collected.')



from typing import (
    Optional,
    Type,
)

from django.db.models import (
    Model,
)

from factory_bo.base import (
    DefaultManagerFactory,
)


class ExistingFactory:
    """
    Класс существующей фабрики
    """
    def __init__(
        self,
        factory_class,
        import_path: str,
        module_path: str,
        factory_class_name_alias: str = None,
    ):
        self._factory_class = factory_class
        self._factory_class_name_alias = factory_class_name_alias
        self._import_path = import_path
        self._module_path = module_path

    def __repr__(self):
        return (
            f'<{self.__class__.__name__} '
            f'@model_label="{self.factory_model_label}" '
            f'@factory_class_name="{self.factory_class_name}">'
        )

    def __str__(self):
        return self.__repr__()

    @property
    def factory_class(self) -> Type[DefaultManagerFactory]:
        return self._factory_class

    @property
    def factory_class_name_alias(self) -> Optional[str]:
        return self._factory_class_name_alias

    @factory_class_name_alias.setter
    def factory_class_name_alias(
        self,
        value: str,
    ):
        self._factory_class_name_alias = value

    @property
    def factory_class_name(self) -> str:
        """
        Возвращает имя класса фабрики или алиас, если он задан
        """
        return self._factory_class_name_alias or self._factory_class.__name__

    @property
    def factory_model(self) -> Type[Model]:
        return self._factory_class.get_model()

    @property
    def factory_model_label(self) -> str:
        return self._factory_class.get_model_label()

    @property
    def import_path(self):
        return self._import_path

    @property
    def module_path(self):
        return self._module_path

    @property
    def get_or_create_fields(self):
        return self._factory_class.get_django_get_or_create()

    @property
    def auto_updated_fields(self):
        return self._factory_class.get_auto_updated_fields()

    @property
    def allow_use_id_field(self):
        return self._factory_class.get_allow_use_id_field()

    @property
    def excluded_fk_fields(self):
        return self._factory_class.get_excluded_fk_fields()



import json
import os
from abc import (
    abstractmethod,
)
from collections import (
    defaultdict,
)
from datetime import (
    datetime,
)
from itertools import (
    groupby,
)
from operator import (
    itemgetter,
)
from typing import (
    Iterable,
    List,
    Optional,
    Set,
)

from django.conf import (
    settings,
)
from isort.api import (
    sort_code_string,
)

from factory_bo.consts import (
    FACTORY_CLASS_NAME_DELIMETER,
    IMPORT_TEMPLATE,
    IMPORT_WITH_ALIAS_TEMPLATE,
    MODEL_CLASS_NAME_ID_DELIMITER,
    TAB_STR,
)
from factory_bo.enums import (
    PreparingModelRecordTagEnum,
    UsingLibraryEnum,
)
from factory_bo.helpers import (
    get_pk_slug,
    is_pseudo_fk_field,
)
from factory_bo.mixins import (
    BaseExporter,
    PreparingModelSortMixin,
)
from factory_bo.records import (
    PreparingModelRecord,
)
from factory_bo.storages import (
    ExistingFactoryStorage,
    ModelFactoryCorrelationStorage,
    PreparingModelRecordStorage,
)
from factory_bo.strings import (
    PYTHON_FIXTURE_PATH_HAVE_NOT_PY_EXTENSION_ERROR,
)


class PreparingModelRecordStorageExporter(BaseExporter):
    """
    Базовый класс экспортев хранилища в файл
    """

    def __init__(
        self,
        preparing_model_record_storage: PreparingModelRecordStorage,
        existing_factory_storage: ExistingFactoryStorage,
    ):
        self._preparing_model_record_storage = preparing_model_record_storage
        self._existing_factory_storage = existing_factory_storage

    @abstractmethod
    def to_string(
        self,
        *args,
        **kwargs,
    ) -> str:
        pass


class BasePyExporter(BaseExporter):
    """
    Базовый Python-экспортер
    """

    @abstractmethod
    def _get_function_declaration(self) -> str:
        """
        Возвращает строку объявления функции
        """
        pass

    def _prepare_libraries_imports(
        self,
        content: str,
    ) -> List[str]:
        """
        Подготавливает импорты библиотек
        """
        libraries_imports = []

        for key_word in UsingLibraryEnum.values.keys():
            if key_word in content:
                libraries_imports.append(
                    UsingLibraryEnum.values[key_word]
                )

        return libraries_imports

    def _prepare_imports(
        self,
        content: str,
    ) -> List[str]:
        libraries_imports = self._prepare_libraries_imports(
            content=content,
        )

        imports = [
            *libraries_imports,
        ]

        return imports

    def _add_imports(
        self,
        content: str,
    ) -> str:
        """
        Добавление и сортировка импортов
        """
        imports = self._prepare_imports(
            content=content,
        )

        content_with_imports = '\n'.join(
            [
                *imports,
                content,
            ]
        )

        return content_with_imports

    def _sort_imports(
        self,
        content: str,
    ):
        """
        Сортировка импортов при помощи isort
        """
        return sort_code_string(
            code=content,
            config=settings.ISORT_CONFIG,
        )

    def _prepare_function_components(self) -> List[str]:
        """
        Подготавливает составляющие функции для дальнейшей конкатинации
        """
        components = [
            self._get_function_declaration(),
        ]

        return components

    def _prepare_function_loader_content(self) -> str:
        """
        Генерирует исходный код функции загрузки предыстории
        """
        components = self._prepare_function_components()

        pre_content = []

        for component in components:
            if component:
                pre_content.extend([component, ''])

        content = '\n'.join(pre_content)

        return content

    def to_string(
        self,
        *args,
        **kwargs,
    ) -> str:
        """
        Возвращает строку содержащую функцию factory_loader с фабриками в тебе
        и импортами всех необходимых фабрик и классов сторонних библиотек
        """
        content = self._prepare_function_loader_content()

        content = self._add_imports(
            content=content,
        )

        content = self._sort_imports(
            content=content,
        )

        return content

    def to_file(
        self,
        *args,
        file_path: Optional[str] = None,
        **kwargs,
    ) -> Optional[str]:
        """
        Формирование Python-фикстуры
        """
        if not file_path.endswith('.py'):
            raise ValueError(
                PYTHON_FIXTURE_PATH_HAVE_NOT_PY_EXTENSION_ERROR.format(
                    fixture_path=file_path,
                )
            )

        return super().to_file(
            *args,
            file_path=file_path,
            **kwargs,
        )


class PyFactoryExporter(
    BasePyExporter,
    PreparingModelSortMixin,
    PreparingModelRecordStorageExporter,
):
    """
    Экспортер фабрик в виде Python-фикстуры
    """

    def __init__(
        self,
        model_factory_correlation_storage: ModelFactoryCorrelationStorage,
        *args,
        **kwargs,
    ):
        self._model_factory_correlation_storage = model_factory_correlation_storage
        self._factory_imports = set()
        self._already_existing_records = []

        super().__init__(*args, **kwargs)

    def _check_getting_record_dependency(
        self,
        preparing_record: PreparingModelRecord,
    ):
        """
        Проверяет, является ли запись только для выборки и есть ли от нее
        зависимые
        """
        return (
            preparing_record.has_only_getting_tag and
            not preparing_record.has_dependence
        )

    def _get_preparing_factories(
        self,
        sorted_preparing_record_model_ids: List[str],
    ) -> str:
        """
        Получение обрабатываемых записей - получаемых и создаваемых в строковом
        представлении
        """
        preparing_factory_strs = []
        deferred_save_fields = defaultdict(list)

        for record_model_id in sorted_preparing_record_model_ids:
            model_label, model_record_pk = record_model_id.split(MODEL_CLASS_NAME_ID_DELIMITER)

            preparing_record = self._preparing_model_record_storage.records.get(
                (
                    model_label,
                    model_record_pk,
                )
            )

            record_fields = {}
            existing_factory = self._model_factory_correlation_storage.get_existing_factory(
                model_label=preparing_record.model_record.model_label,
            )

            # Пропустим получаемые(и только получаемые) записи, для которых
            # нет зависимой записи, чтобы не формировать неиспользуемые
            # переменные и фабрики в фикстуре
            if self._check_getting_record_dependency(preparing_record):
                continue

            self._factory_imports.add(
                (
                    existing_factory.import_path,
                    existing_factory.factory_class_name,
                )
            )

            if preparing_record.has_creating_tag:
                field_names = preparing_record.model_record.fields.keys()
            elif preparing_record.has_getting_tag:
                field_names = existing_factory.get_or_create_fields
            else:
                field_names = []

            if (
                existing_factory.allow_use_id_field and
                'id' in preparing_record.model_record.all_fields
            ):
                field_names = preparing_record.model_record.all_fields.keys()
                model_record_fields = preparing_record.model_record.all_fields
            else:
                model_record_fields = preparing_record.model_record.fields

            existing_factory_class = existing_factory.factory_class_name

            if preparing_record.has_getting_tag:
                variable_name = f'{existing_factory_class.lower()}_{get_pk_slug(preparing_record.model_record.pk)}'

            for field_name in field_names:
                if field_name in existing_factory.auto_updated_fields:
                    continue

                field_value = model_record_fields[field_name]

                if isinstance(field_value, PreparingModelRecord):
                    fk_existing_factory = self._model_factory_correlation_storage.get_existing_factory(
                        model_label=field_value.model_record.model_label,
                    )

                    self._factory_imports.add(
                        (
                            fk_existing_factory.import_path,
                            fk_existing_factory.factory_class_name,
                        )
                    )

                    field_value = (
                        f'{fk_existing_factory.factory_class_name.lower()}_'
                        f'{get_pk_slug(field_value.model_record.pk)}'
                    )

                    # Сохраняем поля для отложенного сохранения после добавления
                    # связанных объектов.
                    if preparing_record.has_getting_tag:
                        if field_name in existing_factory.excluded_fk_fields:
                            deferred_save_fields[field_value].append(
                                (
                                    variable_name,
                                    field_name,
                                )
                            )
                            continue

                    try:
                        # Исключим поля CompositeForeignKey
                        from compositefk.fields import (
                            CompositeForeignKey,
                        )
                        field_obj = preparing_record.model_record.model_class._meta.get_field(field_name)

                        if not isinstance(field_obj, CompositeForeignKey):
                            field_value = f'{field_value}.pk'
                    except ImportError:
                        field_value = f'{field_value}.pk'

                record_fields[field_name] = field_value

            formatted_fields = ''.join(
                [
                    f"{TAB_STR * 2}{k}={v},\n"
                    for k, v in record_fields.items()
                ]
            )

            if preparing_record.has_getting_tag:
                record_str = (
                    f'{TAB_STR}{variable_name} = {existing_factory_class}(\n'
                    f'{formatted_fields}{TAB_STR})'
                )
            else:
                record_str = (
                    f'{TAB_STR}{existing_factory_class}(\n'
                    f'{formatted_fields}{TAB_STR})'
                )
            if record_str not in self._already_existing_records:
                preparing_factory_strs.append(record_str)
                self._already_existing_records.append(record_str)

            # Проверяем нет ли связанных объектов с полями отложенного
            # сохранения, если есть, то добавляем в них и сохраняем.
            if preparing_record.has_getting_tag:
                related_objects = deferred_save_fields.get(variable_name)
                if related_objects:
                    for deferred_object, object_fields in groupby(
                        related_objects,
                        key=itemgetter(0)
                    ):
                        deferred_fields = []
                        for _, field in object_fields:
                            deferred_fields.append(
                                f'{TAB_STR}{deferred_object}.{field} = '
                                f'{variable_name}.pk'
                            )
                        deferred_fields.append(
                            f'{TAB_STR}{deferred_object}.save()'
                        )

                        preparing_factory_strs.append(
                            '\n'.join(deferred_fields)
                        )

        return '\n'.join(preparing_factory_strs)

    def _get_getting_factories(self) -> str:
        """
        Получение всех запрашиваемых записей фабрик в строковом представлении
        """
        getting_records: List[PreparingModelRecord] = list(
            filter(
                lambda record: (
                    PreparingModelRecordTagEnum.GETTING in record.tags and
                    PreparingModelRecordTagEnum.CREATING not in record.tags
                ),
                self._preparing_model_record_storage.records.values()
            )
        )

        sorted_getting_record_model_ids = self._sort_records(
            records=getting_records,
        )

        return self._get_preparing_factories(
            sorted_preparing_record_model_ids=sorted_getting_record_model_ids,
        )

    def _get_creating_factories(self) -> str:
        """
        Получение всех создаваемых записей фабрик в строковом представлении
        """
        creating_records: List[PreparingModelRecord] = list(
            filter(
                lambda record: (
                    PreparingModelRecordTagEnum.CREATING in record.tags
                ),
                self._preparing_model_record_storage.records.values()
            )
        )

        sorted_creating_record_model_ids = self._sort_records(
            records=creating_records,
        )

        return self._get_preparing_factories(
            sorted_preparing_record_model_ids=sorted_creating_record_model_ids,
        )

    def _get_updating_factories(self) -> str:
        """
        Получение всех обновляемых записей фабрик в строковом представлении
        """
        updating_records: List[PreparingModelRecord] = list(
            filter(
                lambda record: (
                    PreparingModelRecordTagEnum.UPDATING in record.tags
                ),
                self._preparing_model_record_storage.records.values()
            )
        )

        sorted_updating_record_model_ids = self._sort_records(
            records=updating_records,
        )

        updating_factory_strs = []

        for record_model_id in sorted_updating_record_model_ids:
            model_label, model_record_pk = record_model_id.split(MODEL_CLASS_NAME_ID_DELIMITER)
            preparing_record = self._preparing_model_record_storage.records.get(
                (
                    model_label,
                    model_record_pk,
                )
            )

            if PreparingModelRecordTagEnum.UPDATING not in preparing_record.tags:
                # такая запись могла попасть сюда после сортировки, которая
                # добавляет в обрабатываемый список записей зависимости
                continue

            record_fields = []
            existing_factory = self._model_factory_correlation_storage.get_existing_factory(
                model_label=preparing_record.model_record.model_label,
            )
            self._factory_imports.add(
                (
                    existing_factory.import_path,
                    existing_factory.factory_class_name,
                )
            )

            for field_name, field_value in preparing_record.changed_fields.items():
                if isinstance(field_value, PreparingModelRecord):
                    fk_existing_factory = self._model_factory_correlation_storage.get_existing_factory(
                        model_label=field_value.model_record.model_label,
                    )

                    self._factory_imports.add(
                        (
                            fk_existing_factory.import_path,
                            fk_existing_factory.factory_class_name,
                        )
                    )

                    field_value = (
                        f'{fk_existing_factory.factory_class_name.lower()}_'
                        f'{get_pk_slug(field_value.model_record.pk)}'
                    )

                    if is_pseudo_fk_field(field_name):
                        field_value = f'{field_value}.id'

                record_fields.append(
                    (
                        field_name,
                        field_value,
                    )
                )

            existing_factory_class = existing_factory.factory_class_name

            variable_name = (
                f'{existing_factory_class.lower()}_'
                f'{get_pk_slug(preparing_record.model_record.pk)}'
            )

            formatted_fields = ''.join(
                [
                    f"{TAB_STR}{variable_name}.{k} = {v}\n"
                    for (k, v) in record_fields
                ]
            )

            record_str = (
                f'{formatted_fields}'
                f'{TAB_STR}{variable_name}.save()'
            )

            updating_factory_strs.append(record_str)

        return '\n'.join(updating_factory_strs)

    def _get_deleting_factories(self) -> str:
        """
        Получение всех удаляемых записей фабрик в строковом представлении
        """
        deleting_records: List[PreparingModelRecord] = list(
            filter(
                lambda record: (
                    PreparingModelRecordTagEnum.DELETING in record.tags
                ),
                self._preparing_model_record_storage.records.values()
            )
        )

        sorted_deleting_record_model_ids = self._sort_records(
            records=deleting_records,
            is_reversed=False,
        )

        deleting_factory_strs = []

        for record_model_id in sorted_deleting_record_model_ids:
            model_label, model_record_pk = record_model_id.split(MODEL_CLASS_NAME_ID_DELIMITER)
            preparing_record = self._preparing_model_record_storage.records.get(
                (
                    model_label,
                    model_record_pk,
                )
            )

            if not preparing_record.has_deleting_tag:
                continue

            existing_factory = self._model_factory_correlation_storage.get_existing_factory(
                model_label=preparing_record.model_record.model_label,
            )

            existing_factory_class_name = existing_factory.factory_class_name

            self._factory_imports.add(
                (
                    existing_factory.import_path,
                    existing_factory_class_name,
                )
            )

            variable_name = (
                f'{existing_factory_class_name.lower()}_'
                f'{get_pk_slug(preparing_record.model_record.pk)}'
            )

            record_str = f'{TAB_STR}{variable_name}.delete()'

            deleting_factory_strs.append(record_str)

        return '\n'.join(deleting_factory_strs)

    def _get_function_declaration(self) -> str:
        """
        Возвращает строку объявления функции загрузки
        """
        return 'def factory_loader(context):'

    def _prepare_factory_imports(self) -> List[str]:
        """
        Подготовливает импорты фабрик
        """
        factory_imports = []

        for import_path, factory_class_name in self._factory_imports:
            if FACTORY_CLASS_NAME_DELIMETER in factory_class_name:
                factory_class_name_alias = factory_class_name
                factory_class_name = factory_class_name.split(FACTORY_CLASS_NAME_DELIMETER)[0]

                import_string = IMPORT_WITH_ALIAS_TEMPLATE.format(
                    import_path=import_path,
                    class_name=factory_class_name,
                    class_name_alias=factory_class_name_alias,
                )
            else:
                import_string = IMPORT_TEMPLATE.format(
                    import_path=import_path,
                    class_name=factory_class_name,
                )

            factory_imports.append(import_string)

        return factory_imports

    def _prepare_imports(
        self,
        content: str,
    ) -> List[str]:
        factory_imports = self._prepare_factory_imports()

        parent_imports = super()._prepare_imports(content)

        imports = [
            *factory_imports,
            *parent_imports,
        ]

        return imports

    def _prepare_function_components(self) -> List[str]:
        """
        Подготавливает составляющие функции для дальнейшей конкатинации
        """
        components = [
            self._get_function_declaration(),
            self._get_getting_factories(),
            self._get_creating_factories(),
            self._get_updating_factories(),
            self._get_deleting_factories(),
        ]

        return components

    def _prepare_file_path(self) -> str:
        """
        Формирует абсолютный путь будущей Python-фикстуры
        """
        now_str = datetime.now().strftime("%Y_%m_%d__%H_%M_%S")
        result_dir_name = 'diff_result'

        result_file_name = f'diff_fixtures__{now_str}.py'

        result_dir_path = os.path.join(
            settings.FACTORY_BO__FIXTURES_DIR_PATH,
            result_dir_name
        )
        if not os.path.exists(result_dir_path):
            os.mkdir(result_dir_path)

        file_path = os.path.join(
            result_dir_path,
            result_file_name
        )

        return file_path

    def to_file(
        self,
        *args,
        file_path: Optional[str] = None,
        **kwargs,
    ) -> Optional[str]:
        """
        Формирование Python-фикстуры
        """
        if not file_path:
            file_path = self._prepare_file_path()

        return super().to_file(
            *args,
            file_path=file_path,
            **kwargs,
        )


class JSONModelExporter(PreparingModelRecordStorageExporter):
    """
    Экспортер обрабатываемых записей хранилища в формате JSON
    """

    def to_string(
        self,
        *args,
        exporting_model_labels: Optional[Set[str]] = None,
        **kwargs,
    ) -> str:
        """
        Вывод обрабатываемых создаваемых записей хранилища в виде JSON-строки
        """
        model_records = defaultdict(list)

        for (model_label, pk), preparing_model_record in self._preparing_model_record_storage.records.items():
            # В результирующий JSON должны попасть записи интересующих моделей
            if exporting_model_labels and model_label not in exporting_model_labels:
                continue

            if PreparingModelRecordTagEnum.CREATING in preparing_model_record.tags:
                model_records[model_label].append(
                    preparing_model_record.model_record.all_fields
                )

        return json.dumps(model_records) if model_records else None


class HierarchyForRecordsPyFactoryExporter(PyFactoryExporter):
    """
    Экспортер фабрик в виде Python-фикстуры для переданных записей моделей.
    """

    def _check_getting_record_dependency(
        self,
        preparing_record: PreparingModelRecord,
    ):
        """Всегда возвращает False, определяя запись как требующую создания переменной."""

        return False

    def to_string(
        self,
        preparing_model_records: Iterable[PreparingModelRecord],
        with_imports: Optional[bool] = False,
        *args,
        **kwargs,
    ) -> str:
        """Формирует python-код с фабриками на основе переданных моделей.

        Args:
            preparing_model_records: Итерируемый объект с моделями.
            with_imports: Добавлять ли в строку импорты.

        Returns:
            Python-код с фабриками.

        """
        sorted_record_model_ids = self._sort_records(preparing_model_records)

        factory_records_string = self._get_preparing_factories(
            sorted_preparing_record_model_ids=sorted_record_model_ids,
        )

        result_string = ''
        for row in factory_records_string.split('\n'):
            result_string += f'{row.removeprefix(" " * 4)}\n'

        if with_imports:
            imports_string = self._add_imports(content='')
            imports_string = self._sort_imports(content=imports_string)
            result_string = f'{imports_string}\n\n{result_string}'

        return result_string



import os

from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
)

from factory_bo.storages import (
    ExistingFactoryStorage,
)


class Command(BaseCommand):
    """
    Команда для поиска использования запрещенных полей наборе
    django_get_or_create
    """
    def handle(self, *args, **options):
        existing_factory_storage = ExistingFactoryStorage()

        default_model_factories_map = (
            existing_factory_storage.get_model_default_factory_map()
        )

        with open(os.path.join(settings.FACTORY_BO__FIXTURES_DIR_PATH, 'wrong_factory_fields.csv'), 'w') as f:  # noqa
            for model_label, existing_factory in default_model_factories_map.values():  # noqa
                get_or_create_fields = set(
                    existing_factory.factory_class.get_django_get_or_create()
                )

                model_fields = (
                    existing_factory_storage.model_storage.get_allowed_fields(
                        model_label=model_label,
                    )
                )

                for field in get_or_create_fields:
                    if (
                        field not in model_fields
                    ):
                        f.write(
                            f'{existing_factory.factory_class_name},{field}\n'  # noqa
                        )



import os
from abc import (
    ABCMeta,
    abstractmethod,
)
from typing import (
    Iterable,
    List,
    Optional,
)

from factory_bo.helpers import (
    topological_sort,
)

from factory_bo.records import (
    PreparingModelRecord,
)


class BaseExporter(metaclass=ABCMeta):
    """
    Абстрактный класс создания экспортеров
    """

    @abstractmethod
    def to_string(
        self,
        *args,
        **kwargs,
    ) -> str:
        pass

    def to_file(
        self,
        *args,
        file_path: Optional[str] = None,
        **kwargs,
    ) -> Optional[str]:
        """
        Вывод результата экспорта в файл
        """
        if not file_path:
            raise ValueError()

        content = self.to_string(*args, **kwargs)

        if content:
            dir_path = os.path.split(file_path)[0]

            os.makedirs(
                dir_path,
                exist_ok=True,
            )

            with open(file_path, 'w') as f:
                f.write(content)
        else:
            print(
                f'File with path "{file_path}" can not create with empty '
                f'content!'
            )

            file_path = ''

        return file_path


class PreparingModelSortMixin:
    """
    Миксин, добавляющий функциональность сортировки обрабатываемых записей по
    степени зависимости между ними
    """

    def _sort_records(
        self,
        records: Iterable[PreparingModelRecord],
        is_reversed: bool = True,
    ) -> List[str]:
        """
        Сортировка обрабатываемых записей
        """
        model_record_id_pairs = set()

        all_model_record_ids = {
            record.model_record.model_label_with_pk
            for record in records
        }

        processed_model_record_ids = set()

        for record in records:
            fk_field_names = set(record.model_record.fk_field_names)
            existing_factory = self._model_factory_correlation_storage.get_existing_factory(
                record.model_record.model_label,
            )
            excluded_fk_fields = existing_factory.excluded_fk_fields

            if excluded_fk_fields:
                fk_field_names = fk_field_names.difference(excluded_fk_fields)

            for field_name in fk_field_names:
                fk_field_value = record.model_record.fields.get(field_name)

                if (
                    fk_field_value and
                    isinstance(fk_field_value, PreparingModelRecord)
                ):
                    model_record_id = record.model_record.model_label_with_pk
                    fk_model_record_id = fk_field_value.model_record.model_label_with_pk

                    if fk_model_record_id not in all_model_record_ids:
                        # Если связанной записи нет в сортируемом изначальном наборе записей, пропускаем
                        continue

                    model_record_id_pairs.add(
                        (
                            model_record_id,
                            fk_model_record_id,
                        )
                    )

                    processed_model_record_ids.add(model_record_id)
                    processed_model_record_ids.add(fk_model_record_id)

        model_record_ids_without_relations = all_model_record_ids.difference(processed_model_record_ids)

        sorting_result = topological_sort(model_record_id_pairs)

        sorted_model_record_ids = (
            list(model_record_ids_without_relations) +
            sorting_result.cyclic +
            list(
                reversed(sorting_result.sorted) if
                is_reversed else
                sorting_result.sorted
            )
        )

        return sorted_model_record_ids



import os
from collections import (
    defaultdict,
)
from typing import (
    Dict,
    List,
    Set,
    Tuple,
)

from django.apps import (
    AppConfig,
    apps,
)
from django.conf import (
    settings,
)
from django.core.management import (
    BaseCommand,
)
from django.db.models import (
    Model,
)
from isort.api import (
    sort_code_string,
)

from factory_bo.consts import (
    TAB_STR,
)
from factory_bo.storages import (
    ExistingFactoryStorage,
)
from factory_bo.strings import (
    FACTORY_CLASS_STR,
    IMPORT_DEFAULT_MANAGER_FACTORY_STR,
)


class Command(BaseCommand):
    """
    Команда для генерации фабрик для моделей, у которых отсутствует дефолтная
    фабрика

    Для моделей, не относящихся к проектам, фабрики будут генерироваться в
    директории

    FACTORY_BO__FIXTURES_DIR_PATH/features/factories/{app_label}.py
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '--apps_list',
            action='store',
            dest='apps_list',
            help=(
                'List of apps to generate factories.'
            )
        )
        parser.add_argument(
            '--exclude_apps',
            action='store',
            dest='exclude_apps',
            help=(
                'Exclude apps'
            )
        )

    def _generate_factory_class_name(
        self,
        model: Model,
    ) -> str:
        """
        Генерация имени фабрики для модели
        """
        model_class_name = model._meta.label.split('.')[1]

        model_class_name_parts = list(
            map(
                lambda part: (
                    part.title() if
                    part.islower() else
                    part
                ),
                model_class_name.split('_')
            )
        )

        return f'Generated{"".join(model_class_name_parts)}Factory'

    def _write_import_web_bb_model_factory(
        self,
        content: str,
        generated_entities: List[str]
    ):
        """
        Запись в файл импортов базовых классов фабрик
        """
        imports_map = {
            'DefaultManagerFactory': IMPORT_DEFAULT_MANAGER_FACTORY_STR,
        }

        for class_name, import_path in imports_map.items():
            if (
                class_name not in content and
                any(class_name in entity for entity in generated_entities)
            ):
                generated_entities.insert(0, import_path)

    def _check_or_create_factories_module(
        self,
        app_label: str,
        app_path: str,
    ) -> Tuple[str, str]:
        """
        Метод создания модуля для фабрик.

        Если модуль создан, то ничего делать не надо.
        """
        # Если модель не из проекта, то выгрузим фабрики отдельно для
        # дальнейшего распределения руками
        if 'web-bb-' not in app_path and 'web_bb_' not in app_path:
            app_path = settings.FACTORY_BO__FIXTURES_DIR_PATH

        features_path = os.path.join(app_path, 'features')
        factories_path = os.path.join(features_path, 'factories')

        os.makedirs(
            name=factories_path,
            exist_ok=True,
        )

        for path in (features_path, factories_path):
            init_py_path = os.path.join(path, '__init__.py')
            if not os.path.exists(init_py_path):
                os.mknod(
                    path=init_py_path,
                    mode=0o777,
                )

        factory_module_path = os.path.join(factories_path, f'{app_label}.py')
        if os.path.exists(factory_module_path):
            with open(factory_module_path) as f:
                content = f.read()
        else:
            content = ''

        return factory_module_path, content

    def _sort_imports(
        self,
        content: str,
    ):
        """
        Сортировка импортов
        """
        return sort_code_string(
            code=content,
            config=settings.ISORT_CONFIG,
        )

    def _get_model_get_or_create_fields(
        self,
        model: Model,
    ) -> str:
        """
        Получение списка полей django_get_or_create в виде строки с
        перечислением через запятую
        """
        model_field_names = [
            field.name
            for field in model._meta.fields
        ]
        django_get_or_create_fields = set()

        django_get_or_create_fields.update(
            set(model_field_names).intersection(
                settings.FACTORY_BO__PSEUDO_SELF_FK_IDS
            )
        )

        django_get_or_create_fields.update(
            set(model_field_names).intersection(
                settings.FACTORY_BO__PSEUDO_FK_IDS
            )
        )

        django_get_or_create_fields_str = f'{TAB_STR*2}#'
        if django_get_or_create_fields:
            django_get_or_create_fields_str = ',\n'.join(
                sorted(
                    map(
                        lambda field_str: f'{TAB_STR*2}# {TAB_STR}\'{field_str}\'',  # noqa
                        django_get_or_create_fields
                    )
                )
            )
            django_get_or_create_fields_str = (
                f'{django_get_or_create_fields_str},'
            )

        return django_get_or_create_fields_str

    def _generate_factory_for_model(
        self,
        grouped_models_by_apps: Dict[Tuple[str, str], Set[Model]],
    ):
        for (app_label, app_path), models in grouped_models_by_apps.items():
            factory_module_path, content = (
                self._check_or_create_factories_module(
                    app_label=app_label,
                    app_path=app_path,
                )
            )

            generated_entities = []

            self._write_import_web_bb_model_factory(
                content=content,
                generated_entities=generated_entities,
            )
            for model in models:
                django_get_or_create_str = (
                    self._get_model_get_or_create_fields(
                        model=model,
                    )
                )

                factory_class_name = self._generate_factory_class_name(
                    model=model,
                )

                base_factory_class = 'DefaultManagerFactory'

                generated_entities.append(
                    FACTORY_CLASS_STR.format(
                        base_factory_class=base_factory_class,
                        factory_class_name=factory_class_name,
                        model_label=model._meta.label,
                        django_get_or_create_str=django_get_or_create_str,
                    )
                )

            self._write_import_web_bb_model_factory(
                content=content,
                generated_entities=generated_entities,
            )

            content = ''.join([content, *generated_entities])

            content = self._sort_imports(
                content=content,
            )

            with open(factory_module_path, 'w') as f:
                f.write(content)

    def _get_apps_configs(self, options: Dict) -> List[AppConfig]:
        """Получение списка AppConfig приложений в соответствии с переданными аргументами.

        Args:
            options: Словарь аргументов командной строки.

        Returns:
            Список AppConfig.

        Raises:
            ValueError: Если передано неверное название приложения.
        """
        apps_configs = []

        if options['apps_list']:
            apps_list = list(filter(None, options.get('apps_list', '').split(',')))
            for app_name in apps_list:
                try:
                    app_config = apps.app_configs[app_name]
                except KeyError as e:
                    raise ValueError(f'No app found with name "{app_name}"') from e
                else:
                    apps_configs.append(app_config)
        else:
            exclude_apps = list(filter(None, options.get('exclude_apps', '').split(',')))
            exclude_apps.extend(settings.FACTORY_BO__EXCLUDED_APPS)
            for app_name, app_config in apps.app_configs.items():
                if app_name in exclude_apps:
                    continue
                apps_configs.append(app_config)

        return apps_configs

    def handle(self, *args, **options):
        apps_configs = self._get_apps_configs(options)

        models = []
        for app_config in apps_configs:
            models.extend(app_config.get_models(include_auto_created=True))

        existing_factory_storage = ExistingFactoryStorage(
            raise_on_nonexistent=False,
        )

        grouped_models_by_apps = defaultdict(set)

        for model in models:
            model_label = model._meta.label
            app_label = model._meta.app_label
            app_path = model._meta.app_config.path

            if (
                model_label in settings.FACTORY_BO__EXCLUDED_MODELS or
                existing_factory_storage.model_factories_map.get(model_label, {}) or
                model._meta.proxy
            ):
                continue

            grouped_models_by_apps[(app_label, app_path)].add(model)

        self._generate_factory_for_model(
            grouped_models_by_apps=grouped_models_by_apps,
        )



from django.conf import (
    settings,
)
from django.core.exceptions import (
    ValidationError,
)
from django.core.management import (
    BaseCommand,
)
from django.db.models import (
    Count,
)

from factory_bo.storages import (
    ExistingFactoryStorage,
)


class Command(BaseCommand):
    """
    Команда для проверки набора полей django_get_or_create на предмет
    уникальной выборки записей
    """

    def handle(self, *args, **options):
        existing_factory_storage = ExistingFactoryStorage()

        project = 'web-bb-core'
        file_path = (
            f'{settings.FACTORY_BO__FIXTURES_DIR_PATH}/'
            f'{project.replace("-", "_")}_need_check_generated_actual_factories.csv'  # noqa
        )
        with open(file_path, 'r') as generated_file:
            factory_class_names = list(
                filter(
                    lambda x: x and 'ForCopy' not in x,
                    [
                        item.split(',')[1].replace('\n', '').replace('Generated', '')  # noqa
                        for item in generated_file.readlines()
                    ]
                )
            )

            for factory_class_name in factory_class_names:
                existing_factory = existing_factory_storage.get_factory_class_by_name(  # noqa
                    factory_class_name=factory_class_name,
                )
                if not existing_factory:
                    raise ValidationError(
                        f'Factory class with name "{factory_class_name}" not '
                        f'found!'
                    )

                model = existing_factory.factory_class.get_model()

                model.objects = model._base_manager

                primary_key = model._meta.pk.column

                duplicate_records = model.objects.values(
                    *existing_factory.get_or_create_fields
                ).annotate(
                    duplicates_count=Count(primary_key),
                ).filter(
                    duplicates_count__gt=1,
                )

                if duplicate_records.exists():
                    print(
                        f'\n\nDuplicate records found! Factory class name - '
                        f'{factory_class_name}, django_get_or_create - '
                        f'({", ".join(existing_factory.get_or_create_fields)})'
                        f'\n\n{duplicate_records.query}'
                    )
